#include "GM.h"
#include "JG.h"
#include "IJGP_Sampler.h"
#include "SampleSearch.h"
#include "Graph.h"
#include "AOTree.h"
#include "AO_IJGP_Sampler.h"
#include <cmath>
#include <cstdlib>
#include <getopt.h>
#include "HyperGraph.h"
#include "Agressive_AO_Sampling.cpp"
#include "AO_Graph_sampler.h"

using namespace std;

typedef enum
{
	MINFILL,TOPOLOGICAL,MINDEGREE
}ORDERING;

char satfile[100]="temp.cnf";

static struct option long_options[]=
{
	{"ordering",required_argument,NULL,'o'},
	{"p-bound",required_argument,NULL,'p'},
	{"i-bound",required_argument,NULL,'i'},
	{"task",required_argument,NULL,'w'},
	{"num-samples",required_argument,NULL,'n'},
	{"time-limit",required_argument,NULL,'t'},
	{"num-iterations",required_argument,NULL,'j'},
	{"input",required_argument,NULL,'f'},
	{"evidence",required_argument,NULL,'e'},
	{"dfs",required_argument,NULL,'d'},
	{"outfile",required_argument,NULL,'g'},
	{"num-components",required_argument,NULL,'c'},
	{"help", no_argument,NULL,'h'},
	{"SS", no_argument,NULL,'s'},
	{"SATFILE",required_argument,NULL,'b'},
	{"aggressive",no_argument,NULL,'@'},
	{"graph",no_argument,NULL,'v'},
	{"ZeroSampler",no_argument,NULL,'z'},
	{"seed",required_argument,NULL,'$'},
};

static const char OPTION_STRING[]="o:i:w:n:t:j:f:hg:p:d:e:c:sb:@:vz";

void print_help()
{
	cout<<"aoijgpss [options] -f inputfile -e evidencefile -d dfs-file -b \n";
	cout<<" --ordering <minfill,topological,mindegree> : What ordering to use\n";
	cout<<" --i-bound <int> : An integer value for the i-bound\n";
	cout<<" --p-bound <int> : An integer value for the p-bound\n";
	cout<<" --num-iterations<int> : The maximum number of iterations for which IJGP is run\n";
	cout<<" --task <int>    : (0) for probability of evidence and (1) for belief updating\n";
	cout<<" --num-samples <int> : The maximum number of samples to be drawn\n";
	cout<<" --time-limt <int> : The maximum time in seconds for which the algorithm is run\n";
	cout<<" --outfile<string> : path where the output file is stored (default out) \n";
	cout<<" --num-components <int> : number of components\n";
	cout<<" -SS : use SampleSearch instead of pure Sampling\n";
	cout<<" --aggressive <int> : Use Aggressive Sampling parameter\n";
	cout<<" --graph : Use AO graph sampling\n";
	cout<<" --ZeroSampler: Run Relational consistency algorithm\n";
	cout<<" --seed<int>: Use this seed\n";
	cout<<" --help : Print this help\n";

}
void readOrder(GM& gm, char*infilename, vector<int>& order, vector<int>& sampling_order)
{
	ifstream in(infilename);
	//vector<int> ids_to_vars;
	//for(int i=0;i<gm.var_to_ids.size();i++)
	//{
	//	if((gm.var_to_ids[i]+1)> ids_to_vars.size())
	//	{
	//		int old_size=ids_to_vars.size();
	//		ids_to_vars.resize(gm.var_to_ids[i]+1);
	//		for(int j=old_size;j<ids_to_vars.size();j++)
	//			ids_to_vars[j]=-1;
	//	}
	//	ids_to_vars[gm.var_to_ids[i]]=i;
	//}
	//vector<bool> check(gm.variables.size());
	////while(in.good())
	//for(int i=0;i<gm.variables.size();i++)
	//{
	//	int temp;
	//	in>>temp;
	//	assert(ids_to_vars[temp]!=-1);
	//	sampling_order.push_back(ids_to_vars[temp]);
	//	check[ids_to_vars[temp]]=true;
	//}
	//for(int i=0;i<gm.variables.size();i++)
	//	if(!check[i])
	//	{
	//		cerr<<"Error in reading order\n";
	//		sampling_order=vector<int>();
	//		in.close();
	//		return;
	//	}
	//
	//cout<<sampling_order.size()<<" "<<gm.variables.size()<<endl;
	//assert(sampling_order.size()==gm.variables.size());
	sampling_order=vector<int>(gm.variables.size());
	order=vector<int> (gm.variables.size());
	int num_vars;
	in>>num_vars;
	assert(num_vars==(int)gm.variables.size());
	for(int i=0;i<gm.variables.size();i++)
		in>>order[i];
	for(int i=0;i<gm.variables.size();i++)
		in>>sampling_order[i];
	in.close();
}
unsigned int myRandom::seed;
int main(int argc, char *argv[])
{

	char ordering[100];
	PseudoTree pseudo_tree;
	int i_bound=3,p_bound=0,w_bound=0,num_samples=-1,time_limit=10,num_iterations=10;
	GM gm;
	int num_components=2;
	char graphfile[100];
	bool writegraph=false;
	bool p_of_e=true;
	int task;
	char outfilename[100]="out";
	vector<int> sampling_order;
	vector<int> order;
	bool sample_search=false;
	bool aggressive=false;
	SAMPLER_TYPE sampler_type=POSITIVE_SAMPLER;
	int ag_value=2;
	unsigned int seed=time(NULL);
	if(argc==1)
	{
		print_help();
		return 1;
	}
	bool graph_sampling=false;
	while(1)
	{
		int option_index =0;
		int opt = getopt_long(argc,argv,OPTION_STRING,long_options,&option_index);
		// Detect end of options
		if (opt == -1)
			break;
		switch(opt)
		{
		case 'o':
			sscanf(optarg,"%s",&ordering);
			break;
		case '$':
			sscanf(optarg,"%ld",&seed);
			break;
		case 'v':
			graph_sampling=true;
			break;
		case 'z':
			sampler_type=ZERO_SAMPLER;
			break;
		case 'p':
			sscanf(optarg,"%ld",&p_bound);
			break;
		case '@':
			aggressive=true;
			sscanf(optarg,"%ld",&ag_value);
			break;
		case 'i':
			sscanf(optarg,"%ld",&i_bound);
			break;
		case 'c':
			sscanf(optarg,"%ld",&num_components);
			break;
		case 'n':
			sscanf(optarg,"%ld",&num_samples);
			break;
		case 't':
			sscanf(optarg,"%ld",&time_limit);
			break;
		case 'w':
			sscanf(optarg,"%d",&task);
			if(task > 0)
				p_of_e=false;
			else
				p_of_e=true;
			break;
		case 'j':
			sscanf(optarg,"%ld",&num_iterations);
			break;
		case 's':
			sample_search=true;
			break;
		case 'b':
			strcpy(optarg,satfile);
			break;
		case 'f':
			gm=GM();
			if(!gm.readCNF(optarg))
			{
			//gm.readVIB(optarg);
				gm.readErgo(optarg);
			}
			
			break;
		case 'e':
			gm.readEvidenceErgo(optarg);
			cout<<"Read evidence\n";
			break;
		case 'd':
			//readSamplingOrder(gm,optarg,sampling_order);
			readOrder(gm,optarg,order,sampling_order);
			//pseudo_tree=PseudoTree(gm,optarg,sampling_order);
			break;
		case 'g':
			strcpy(outfilename,optarg);
			break;
		case 'h':
			print_help();
			return 1;
		default:
			print_help();
			return 1;
		}
	}
	myRandom::seed=seed;
	//gm.reduce(i_bound);
	cerr<<"i-bound= "<<i_bound<<endl;
	//assert(!order.empty());
	//assert(sampling_order.size()==order.size());
	if(order.empty())
	{
		if(sampler_type==1)
		{
			GM csp;
			gm.convertToCN(csp);
			Graph g;
			g.makeGraph(&csp);
			g.getMinFillOrdering(order);
		}
		else
		{
			Graph g;
			g.makeGraph(&gm);
			g.getMinFillOrdering(order);
		}
	}

	JG jg(gm,i_bound,num_iterations,order,LS);

	cerr<<"Join graph constructed\n";

	//Construct graph from the clusters of the join graph
	
	time_t start_time,end_time;
	start_time=time(NULL);
	jg.propagate();
	end_time=time(NULL);



	Graph g;
	jg.getGraph(gm,g);

	//Update order
	// This is an ugly hack
	sampling_order=vector<int>(order.size());
	for(int i=0;i<sampling_order.size();i++)
	{
		//order[sampling_order.size()-i-1]=sampling_order[i];
		//cerr<<sampling_order[i]<<" ";
		sampling_order[i]=order[order.size()-i-1];
	}
	//cerr<<endl;
	// End update order

	
	pseudo_tree=PseudoTree(gm,g,sampling_order);
	// Use hmetis to figure out the ordering
	//pseudo_tree=PseudoTree(gm,jg,sampling_order,order);
	
	if(p_of_e)
	{
		
		Double pe;
		/*if(aggressive)
		{
			AGGRRESSIVE_AO_SAMPLING sampler;
			pe=sampler.computePE(gm,jg,p_bound,order,pseudo_tree,ag_value,time_limit,num_samples);
			pe*=gm.mult_factor;
			ofstream outfile(outfilename,ios::app);
			sampler.stats.printStats(outfile);
			outfile.close();
		}*/
		if(graph_sampling)
		{
			AO_Graph_Sampler sampler;
			sampler.type=sampler_type;
			if(!sample_search)
			{
				pe=sampler.computePE(gm,jg,p_bound,order,pseudo_tree,time_limit,num_samples);
			}
			else
			{
				pe=sampler.computePESS(gm,jg,p_bound,order,pseudo_tree,time_limit,num_samples);
			}
			pe*=gm.mult_factor;
			ofstream outfile(outfilename,ios::app);
			sampler.stats.printStats(outfile);
			outfile.close();
		}
		else if(sample_search)
		{
			AO_IJGP_Sampler sampler;
			sampler.type=sampler_type;
			pe=sampler.computePESS(gm,jg,p_bound,order,pseudo_tree,time_limit,num_samples);
			pe*=gm.mult_factor;
			ofstream outfile(outfilename,ios::app);
			sampler.stats.printStats(outfile);
			outfile.close();
		}
		else
		{
			AO_IJGP_Sampler sampler;
			sampler.type=sampler_type;
			pe=sampler.computePE(gm,jg,p_bound,order,pseudo_tree,time_limit,num_samples);
			pe*=gm.mult_factor;
			ofstream outfile(outfilename,ios::app);
			sampler.stats.printStats(outfile);
			outfile.close();
		}
		/*pe*=gm.mult_factor;
		ofstream outfile(outfilename,ios::app);
		sampler.stats.printStats(outfile);
		outfile.close();
		*/
		
		cout<<"---------------------------RESULT------------------------------\n";
		cout<<"\t Log probability of evidence = "<<log(pe.value())<<endl;
		cout<<"\t Probability of evidence = "<<pe<<endl;
		cout<<"---------------------------RESULT------------------------------\n";

	}
	else
	{
		vector<vector<Double> > beliefs;
		IJGP_Sampler sampler;
		sampler.computeBeliefs(gm,jg,p_bound,order,time_limit,num_samples,beliefs);
		ofstream outfile(outfilename);
		outfile<<beliefs.size()<<endl;
		for(int i=0;i<beliefs.size();i++)
		{
			outfile<<i<<" "<<beliefs[i].size()<<" ";
			for(int j=0;j<beliefs[i].size();j++)
			{
				outfile<<beliefs[i][j]<<" ";
			}
			outfile<<endl;
		}
		outfile.close();
	}

}
