#include <iostream>
#include <fstream>
#include "BE.h"
#include "GM.h"
#include "Graph.h"
#include "RB_IJGP_Sampler.h"
#include "SampleSearch.h"
#include "SampleSearch.cpp"
#include <cmath>

using namespace std;
#include <sys/time.h>
#include <sys/resource.h>
static inline double get_cpu_time(void) 
{
  struct rusage ru;
  getrusage(RUSAGE_SELF, &ru);
  return (double)ru.ru_utime.tv_sec + (double)ru.ru_utime.tv_usec / 1000000; 
}

#define LIMIT 10000000
int main(int argc, char* argv[])
{
	int i_bound,p_bound=0,w_bound=0,num_samples=-1,time_limit=10,num_iterations=10;
	GM gm;
	gm.readUAI08(argv[1]);
       
	//Read evidence
	ifstream in(argv[2]);
	int num_evidence=0;
	vector<int> evidence;
	if(in.good())
	{
		in>>num_evidence;
		for(int i=0;i<num_evidence;i++)
		{
			int var,val;
			in>>var>>val;
			gm.variables[var]->value()=val;
		}
	}
	in.close();   
	cerr<<"num evidence = "<<num_evidence<<endl;
	
	ifstream margfile(argv[3]);
	while(margfile.good()){
	  char id_char;
	  char buffer[500000];
	  margfile.get(id_char);
	  if(id_char != 'm'){
	    margfile.putback(id_char);
	    margfile.getline(buffer);
	    continue;
	  }
	  vector<vector<Double> > marginal(gm.variables.size());
	  for(int i=0;i<gm.variables.size();i++){
	    int curr_dom_size;
	    margfile>>curr_dom_size;
	    if(curr_dom_size != gm.variables[i]->domain_size()){
	      cerr<<"Something Wrong in "<<argv[3]<<endl;
	      margfile.getline(buffer);
	      continue;
	    }
	    marginal[i]=vector<double> (gm.variables[i]->domain_size());
	    for(int j=0;j<gm.variables[i]->domain_size();j++){
	      margfile>>marginal[i][j];
	    }
	  }
	  
	  // Fix the marginal
	  if (gm.mode == DET){
	    gm.convertToSAT();
	    gzFile in = gzopen(satfile,"rb");
	    parse_DIMACS(in,solver);
	    gzclose(in);
	  }
	  
	}
	margfile.close();
	
	cerr<<"Limit = "<<limit<<endl;
	vector<int> new_order;
	gm.rearrangeOrdering(order,clusters,new_order,limit);
	cerr<<"#sampled variables = "<<new_order.size()<<" out of "<<gm.variables.size()<<endl;
	
	//computeIBound(gm,i_bound);
	vector<Variable*> cond_variables(new_order.size());
	for(int i=0;i<new_order.size();i++)
		cond_variables[i]=gm.variables[new_order[i]];
	int num_values=Variable::getDomainSize(cond_variables);
	Double pe;
	cerr<<"num values = "<<num_values<<endl;
	Solver solver;
	
	if(gm.mode==DET){
	  gm.convertToSAT();
	  gzFile in = gzopen(satfile, "rb");
	  parse_DIMACS(in,solver);
	  gzclose(in);
	}
	int n=cond_variables.size();
	int a[n];
	int f[n+1];
	int o[n];
	int m[n];
	// Initialize
	for(int i=0;i<n;i++){
	  cond_variables[i]->value()=0;
	  a[i]=0;
	  f[i]=i;
	  o[i]=1;
	  m[i]=cond_variables[i]->domain_size();
	}
	f[n]=n;
	cout <<"t "<<get_cpu_time()<<endl;
	 
	while (1)
	{
	  bool has_sol=true;
	  if(gm.mode == DET){
	    vec<Lit> assumptions;
	    for(int j=0;j<cond_variables.size();j++){
	      int id=cond_variables[j]->id();
	      int value=cond_variables[j]->value();
	      assumptions.push(Lit(gm.csp_to_sat_variables[id][value]));
	    }
	    if(!solver.solve(assumptions)){
	      has_sol=false;
	    }
	  }
	  if(has_sol){
	    BE be(gm.variables,gm.functions,order);
	    pe+=be.log_pe.toDouble();
	  }
	 
	  Double tmp(pe);
	  tmp*=gm.mult_factor;
	  if(!tmp.isZero())
	    cout<<"z "<<log10(tmp.value())<<endl;
	  int j=f[0];
	  f[0]=0;
	  if(j==n)
	    break;
	  a[j]=a[j]+o[j];
	  cond_variables[j]->value()=a[j];
	  if(a[j]==0 || a[j]==(m[j]-1)){
	    o[j]=-o[j];
	    f[j]=f[j+1];
	    f[j+1]=j+1;
	  }
	}
	
	
	pe*=gm.mult_factor;
	if(!pe.isZero()){
	  cout<<"z "<<log10(pe.value())<<endl;
	}
	cerr<<"---------------------------RESULT------------------------------\n";
	cerr<<"\t Log Base e probability of evidence = "<<log(pe.value())<<endl;
	cerr<<"\t Log Base 10 probability of evidence = "<<log(pe.value())/log((double)10.0)<<endl;
	cerr<<"\t Probability of evidence = "<<pe<<endl;
	cerr<<"---------------------------RESULT------------------------------\n";

	
}
