#include <iostream>
#include <fstream>
#include "BE.h"
#include "GM.h"
#include "Graph.h"
#include "RB_IJGP_Sampler.h"
#include "SampleSearch.h"
#include "JT.h"

using namespace std;
#define IBOUNDLIMIT 32768
#include <sys/time.h>
#include <sys/resource.h>
static inline double get_cpu_time(void) 
{
  struct rusage ru;
  getrusage(RUSAGE_SELF, &ru);
  return (double)ru.ru_utime.tv_sec + (double)ru.ru_utime.tv_usec / 1000000; 
}
void computeIBound(GM& gm, int& i_bound)
{
	int max_domain_size=0;
	for(int i=0;i<gm.variables.size();i++)
	{
		if(max_domain_size < (int) gm.variables[i]->domain_size())
			max_domain_size=gm.variables[i]->domain_size();
	}
	cerr<<"Max dom size = "<<max_domain_size<<endl;
	double m=log((double)IBOUNDLIMIT)/log((double)max_domain_size);
	i_bound=(int)m;
}

int main(int argc, char* argv[])
{
	int i_bound=6,p_bound=0,w_bound=0,num_samples=1000000,time_limit=10,num_iterations=20;
	GM gm;
	gm.readUAI08(argv[1]);
	vector<int> order;
	//Read evidence
	ifstream in(argv[2]);
	int num_evidence=0;
	vector<int> evidence;
	if(in.good())
	{
		in>>num_evidence;
		for(int i=0;i<num_evidence;i++)
		{
			int var,val;
			in>>var>>val;
			gm.variables[var]->value()=val;
			evidence.push_back(var);
		}
	}
	vector<set<int> > clusters;
	double estimate;
	double limit=(double)10000;
	//sscanf(argv[4],"%lf",&limit);
	//sscanf(argv[5],"%d",&i_bound);
	bool run_samplesearch=false;
	//gm.removeIrrelevantNetwork(evidence);
	gm.setEvidenceBeliefsUAI08(evidence);
	if(gm.mode == DET)
	{
		cerr<<"Reducing domains\n";
		gm.reduceDomains();
	}
	computeIBound(gm,i_bound);

	gm.getMinFillOrdering(order,clusters,estimate);

	// Compute the treewidth
	int treewidth=0;
	for(int i=0;i<clusters.size();i++){
	  if(treewidth < (int) clusters[i].size())
	    treewidth=(int) clusters[i].size();
	}
	if(treewidth <= i_bound){
	  JT jt(gm.variables,gm.functions,clusters,order);
	  jt.propagate();
	  gm.printMarginalsUAI08(jt.marginals);
	  return 1;
	}
	cerr<<"num evidence = "<<num_evidence<<endl;
	cerr<<"Treewidth  = "<<treewidth<<endl;
	vector<Function> true_marginals;
	cout <<"t "<<get_cpu_time()<<endl;
	while(1){
	  
	  JG jg (gm,i_bound,num_iterations,order,LSS);
	  jg.propagate();
	  vector<Function> marginals(gm.variables.size());
	  for(int j=0;j<gm.variables.size();j++){
	    vector<Variable*> curr_variable;
	    curr_variable.push_back(gm.variables[j]);
	    bool found=false;
	    for(int i=0;i<jg.nodes.size();i++){
	      vector<Variable*> tmp_vars;
	      do_set_intersection(jg.nodes[i]->variables(),curr_variable,tmp_vars,less_than_comparator_variable);
	      if(!tmp_vars.empty()){
		found=true;
		jg.nodes[i]->getMarginal(curr_variable,marginals[j]);
		break;
	      }  
	    }
	    if(!found){
	      marginals[j].variables()=curr_variable;
	      marginals[j].table()=vector<Double>(gm.variables[j]->domain_size());
	      for(int i=0;i<gm.variables[j]->domain_size();i++){
		marginals[j].table()[i]=Double((double)1.0/(double)gm.variables[j]->domain_size());
	      }
	    }
	  }
	  //cout<<"Marginals generated\n";
	  // Delete the join graph
	  jg.clear();
	  if(true_marginals.empty()){
	    true_marginals=marginals;
	  }
	  else{
	    for(int i=0;i<true_marginals.size();i++){
	      for(int j=0;j<true_marginals[i].table().size();j++){
		true_marginals[i].table()[j]+=marginals[i].table()[j];
	      }
	    }
	  }
	  gm.printMarginalsUAI08(true_marginals);

	  i_bound++;
	  if(i_bound >= treewidth){
	    JT jt(gm.variables,gm.functions,clusters,order);
	    jt.propagate();
	    gm.printMarginalsUAI08(jt.marginals);
	    return 1;
	  }
	}
	
}
