#include <iostream>
#include <fstream>
#include "BE.h"
#include "GM.h"
#include "Graph.h"
#include "RB_IJGP_Sampler.h"


using namespace std;
//#define IBOUNDLIMIT 3000 //65646
//#define LIMIT 1000 //0000
#include <sys/time.h>
#include <sys/resource.h>
static inline double get_cpu_time(void) 
{
  struct rusage ru;
  getrusage(RUSAGE_SELF, &ru);
  return (double)ru.ru_utime.tv_sec + (double)ru.ru_utime.tv_usec / 1000000; 
}

void computeIBound(GM& gm, int& i_bound)
{
	double IBOUNDLIMIT=pow((double)2.0,(double)i_bound);
	int max_domain_size=0;
	for(int i=0;i<gm.variables.size();i++)
	{
		if(max_domain_size < (int) gm.variables[i]->domain_size())
			max_domain_size=gm.variables[i]->domain_size();
	}
	cerr<<"Max dom size = "<<max_domain_size<<endl;
	double m=log((double)IBOUNDLIMIT)/log((double)max_domain_size);
	i_bound=(int)m;
}
// Parameters
int i_bound=12;
int p_bound=0,w_bound=0;
int rb_bound=7;
// We will stop after a million samples
int num_samples=100000000;
// The time limit is ignored, we run forever or until a million samples are drawn
int time_limit=10;
// The number of iterations of IJGP is set to 10
int num_iterations=10;
// The number of restarts is set to 1
int max_restarts=1;
double damp=0.9;
unsigned int seed = 123456789;
string ordering="minfill";
string sampling_ordering="revminfill";
string outfilename="out";
int interval=100;
string uaifilename;
string evidfilename;
vector<int> evidence;
string task="counting";
int help(){
	cerr<<"Format is : ss [options] uaifilename evidfilename\n";
	cerr<<"Setting proper options is very important, otherwise you will end up with bad results\n"; 
	cerr<<"The way you set options is by using --option value\n";
	cerr<<"Possible options\n";
	cerr<<" \t--max-restarts value (default value is 1)\n";
	cerr<<" \t--i-bound value (default value is 12)\n";
	cerr<<" \t--rb-bound value (default value is 7, stands for rao-blackwellised bound)\n";
	cerr<<" \t--ijgp-iter value (default value is 10)\n";
	cerr<<" \t--order [minfill|mindegree|lex|topological] (default value is minfill, ordering used for join graph generation )\n";
	cerr<<" \t--sampling-order[revminfill|revmindegree|lex|topological|minfillCSP] (default value is minfill)\n";
	cerr<<" \t--ijgp-damp value (default value is 0.9)\n";
	cerr<<" \t--seed value (default value is 123456789 )\n";
	cerr<<" \t--task [counting|marginal] (default value is counting)\n";
	cerr<<"The following options are just for storing results\n";
	cerr<<" \t--interval value (default value is 100)\n";
	cerr<<" \t--out filename (default value is out.ss)\n";
	exit(1);
}
void readParameters(int argc, char* argv[])
{
	int optind=1;
   	while (optind < (argc-2)) {
    	string sw = argv[optind];
     	if(sw == "--max-restarts"){
        	optind++;
        	max_restarts=atoi(argv[optind]);
     	}
	else if (sw=="--ijgp-iter"){
	  optind++;
	  num_iterations=atoi(argv[optind]);
	}
     	else if (sw == "--i-bound"){
     		optind++;
     		i_bound=atoi(argv[optind]);
     	}
	else if (sw == "--task"){
	  optind++;
	  task=argv[optind];
	}
     	else if (sw == "--rb-bound"){
     		optind++;
     		rb_bound=atoi(argv[optind]);
     	}
     	else if (sw == "--order"){
     		optind++;
     		ordering=argv[optind];
     	}
		else if (sw == "--sampling-order"){
			optind++;
			sampling_ordering=argv[optind];
		}
     	else if (sw == "--ijgp-damp"){
     		optind++;
     		damp=atof(argv[optind]);		
     	}
     	else if (sw == "--interval"){
     		optind++;
     		interval=atoi(argv[optind]);
     	}
     	else if (sw == "--out"){
     		optind++;
     		outfilename=argv[optind];
     	}
     	else if (sw == "--seed"){
     		optind++;
     		seed=atoi(argv[optind]);
     	}
     	else if (sw == "--help"){
     		help();
     	}
     	else{
     		help();
     	}
    	optind++;
   }
   if (argc < 2){
   	help();
   }
   uaifilename=argv[argc-2];
   evidfilename=argv[argc-1];
}
void readEvidence(GM& gm)
{
	evidence=vector<int>();
	ifstream in(evidfilename.c_str());
	int num_evidence=0;
	if(in.good())
	{
		in>>num_evidence;
		for(int i=0;i<num_evidence;i++)
		{
			int var,val;
			in>>var>>val;
			gm.variables[var]->value()=val;
			evidence.push_back(var);
		}
	}
	in.close();
}
int main(int argc, char* argv[])
{
	//int i_bound=6,p_bound=0,w_bound=0,num_samples=1000000,time_limit=10,num_iterations=10,max_restarts=1;
	readParameters(argc,argv);
	GM gm;
	myRandom::seed=static_cast<unsigned int> (seed);
	myRandom::base.seed((int)myRandom::seed);
	myRandom::genDoubleVal= RealGenerator(myRandom::base,myRandom::real_dist);
	myRandom::genIntVal=IntGenerator(myRandom::base,myRandom::int_dist);

	gm.readUAI08(uaifilename.c_str());
	readEvidence(gm);
	
	vector<int> order;
	vector<int> topo_order;
	double estimate;
	double limit=pow((double)2.0,(double)rb_bound);
	if (gm.type==BAYES && (sampling_ordering=="topological" || ordering=="topological")){
		Graph graph;
		graph.makeDirectedGraph(&gm);
		graph.getTopologicalOrdering(topo_order);
	}
	if (gm.type==MARKOV && (sampling_ordering=="topological" || ordering=="topological")){
		cerr<<"Error: Topological ordering used for a markov network\n";
		exit(-1);
	}
	if (task=="counting"){
		gm.removeIrrelevantNetwork(evidence);
	}else if (task=="marginals"){
		gm.setEvidenceBeliefsUAI08(evidence);
	}
	else{
		cerr<<"Error: Invalid task\n";
		exit(-1);
	}
	
	if(gm.mode == DET)
	{
		gm.reduceDomains();
	}
	//else{
	//cerr<<"Error: SampleSearch not relevant for this problem, no zeros present\n";
	//exit(-1);
	//}
	
	// Normalize the i_bound based on the domain size
	//computeIBound(gm,i_bound);
	order=vector<int>();
	vector<set<int> > clusters;
	estimate=0.0;
	// Set the sampling order
	if(sampling_ordering=="topological"){
		gm.getTopologicalOrdering(order,clusters,estimate,topo_order);
	}
	else if (sampling_ordering=="revminfill"){
		gm.getMinFillOrdering(order,clusters,estimate);
	}
	else if (sampling_ordering == "revmindegree"){
		gm.getMinDegreeOrdering(order,clusters,estimate);
	} 
	else if(gm.mode == DET && sampling_ordering == "minfillCSP"){
		GM csp;
		gm.convertToCN(csp);
		Graph g;
		g.makeGraph(&csp);
		int treewidth = g.getMinFillOrdering(order);
		cerr<<"I-bound = "<<i_bound<<" Treewidth = "<<treewidth<<endl;
		gm.getClusters(order,clusters);
		//gm.rearrangeOrdering(order,clusters,sampling_order,limit);
	}else {
		cerr<<"Error: Wrong Sampling ordering type\n";
		exit(-1);
	}
	vector<int> sampling_order;
	gm.rearrangeOrdering(order,clusters,sampling_order,limit);
	cerr<<"#sampled variables = "<<sampling_order.size()<<" out of "<<gm.variables.size()<<endl;

	
	cout <<"t "<<get_cpu_time()<<endl;
	
	//Join graph construction
	// Set join graph ordering
	vector<int> jg_order;
	vector<set<int > > jg_clusters;
	if(ordering=="topological"){
		for(int i=0;i<gm.copy_of_variables.size();i++){
			int var=topo_order[i];
			if(gm.copy_of_variables[var]->id()==INVALID_VALUE){
				continue;
			}
			jg_order.push_back(gm.copy_of_variables[var]->id());
		}
		assert(jg_order.size()==gm.variables.size());
		//gm.getTopologicalOrdering(jg_order,jg_clusters,estimate,topo_order);
	}
	else if (ordering=="minfill"){
		gm.getMinFillOrdering(jg_order,jg_clusters,estimate);
	}
	else if (ordering == "mindegree"){
		gm.getMinDegreeOrdering(jg_order,jg_clusters,estimate);
	} 
	else {
		cerr<<"Error: Wrong ordering type\n";
		exit(-1);
	}
	
	JG jg(gm,i_bound,num_iterations,jg_order,LSS);
	cerr<<"Join graph constructed\n";
	jg.propagate();

	
	if (task =="counting")
	{
		RB_IJGP_Sampler sampler;
		Double pe=sampler.computePE(gm,jg,p_bound,order,sampling_order,time_limit,num_samples);
	}
	else
	{
		RB_IJGP_Sampler sampler;
		sampler.computeBeliefs(gm,jg,p_bound,clusters,order,sampling_order,time_limit,num_samples);
	}
}
