#include "GM.h"
#include "JG.h"
#include "IJGP_Sampler.h"
#include "SampleSearch.h"
#include "Graph.h"
#include <cmath>
#include <cstdlib>
#include <getopt.h>

using namespace std;
/*************************************************************************************/
#ifdef _MSC_VER
#include <ctime>

static inline double cpuTime(void) {
    return (double)clock() / CLOCKS_PER_SEC; }
#else

#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

static inline double cpuTime(void) {
    struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return (double)ru.ru_utime.tv_sec + (double)ru.ru_utime.tv_usec / 1000000; }
#endif

typedef enum
{
	MINFILL,TOPOLOGICAL,MINDEGREE
}ORDERING;



static struct option long_options[]=
{
	{"outputfile",required_argument,NULL,'o'},
	{"i-bound",required_argument,NULL,'i'},
	{"num-iterations",required_argument,NULL,'j'},
	{"input",required_argument,NULL,'f'},
	{"outfile",required_argument,NULL,'g'},
	{"help", no_argument,NULL,'h'},
	{"evidencefile", no_argument,NULL,'e'},
};

static const char OPTION_STRING[]="o:i:j:f:hg:e:";

void print_help()
{
	cout<<"ijgp [options] -f inputfile -e evidencefile -o outfile\n";
	cout<<" --i-bound <int> : An integer value for the i-bound\n";
	cout<<" --num-iterations<int> : The maximum number of iterations for which IJGP is run\n";
	cout<<" --outfile <string> : The name of the output file (default is out)\n";
	cout<<" --help : Print this help\n";



}

int main(int argc, char *argv[])
{ 
	int i_bound=3,p_bound=0,w_bound=0,num_samples=-1,time_limit=10,num_iterations=10;
	GM gm;
	char outfilename[100]="out";
	char outfile[100];
	bool writegraph=false;
	if(argc==1)
	{
		print_help();
		return 1;
	}
	while(1)
	{

		int option_index =0;
		int opt = getopt_long(argc,argv,OPTION_STRING,long_options,&option_index);
		// Detect end of options
		if (opt == -1)
			break;
		switch(opt)
		{
		case 'o':
			strcpy(outfile,optarg);
			break;
		case 'i':
			sscanf(optarg,"%ld",&i_bound);
			break;
		case 'j':
			sscanf(optarg,"%ld",&num_iterations);
			break;
		case 'f':
			//if(!gm.readCNF(optarg))
			//{
				gm=GM();
				//gm.readVIB(optarg);
				gm.readErgo(optarg);
				//char evidfile[100];
				//sprintf(evidfile,"%s.evid",optarg);
				//gm.readEvidence(evidfile);
				//gm.readEvidenceErgo(evidfile);
			//}
				break;
		case 'e':
			
				gm.readEvidenceErgo(optarg);
			
			break;
		case 'g':
			strcpy(outfilename,optarg);

			break;
		case 'h':
			print_help();
			return 1;
		default:
			print_help();
			return 1;
		}
	}
	vector<int> order(gm.variables.size());

	GM csp;
	gm.convertToCN(csp);
	Graph g;
	g.makeGraph(&csp);
	g.getMinFillOrdering(order);


	gm.reduceBN(NULL,i_bound,order);
	gm.writeErgo(outfile);
}