#include <iostream>
#include <fstream>
#include "BE.h"
#include "GM.h"
#include "Graph.h"
#include "RB_IJGP_Sampler.h"
#include "SampleSearch.h"

using namespace std;

#define IBOUNDLIMIT 1024
#define LIMIT 1000000


#include <sys/time.h>
#include <sys/resource.h>
static inline double get_cpu_time(void) 
{
  struct rusage ru;
  getrusage(RUSAGE_SELF, &ru);
  return (double)ru.ru_utime.tv_sec + (double)ru.ru_utime.tv_usec / 1000000; 
}

void computeIBound(GM& gm, int& i_bound)
{
	int max_domain_size=0;
	for(int i=0;i<gm.variables.size();i++)
	{
		if(max_domain_size < (int) gm.variables[i]->domain_size())
			max_domain_size=gm.variables[i]->domain_size();
	}
	cerr<<"Max dom size = "<<max_domain_size<<endl;
	double m=log((double)IBOUNDLIMIT)/log((double)max_domain_size);
	i_bound=(int)m;
}

int main(int argc, char* argv[])
{
	int i_bound=6,p_bound=0,w_bound=0,num_samples=1000000,time_limit=10,num_iterations=10;
	bool is_wcsp=false;
	GM gm;
	// Random Generation: Start
	myRandom::seed=static_cast<unsigned int> (atoi(argv[3]));
	//myRandom::seed=atoi(argv[3]);
	myRandom::base.seed((int)myRandom::seed);
	myRandom::genDoubleVal= RealGenerator(myRandom::base,myRandom::real_dist);
	myRandom::genIntVal=IntGenerator(myRandom::base,myRandom::int_dist);
	
	// Random Generation: End
	//gm.readUAI08(argv[1]);
	gm.readCNF(argv[1]);
	gm.type=MARKOV;
	gm.mode=DET;
	string str(argv[1]);
	string str2("wcsp");
	size_t found;
	found=str.find(str2);
	//if((int)gm.wcsp_probs.size() <= 9 && gm.mode==POSITIVE)
	if(found!=string::npos && gm.mode==POSITIVE)
	  is_wcsp=true;
	
	for(int i=0;i<gm.wcsp_probs.size();i++)
	  cerr<<gm.wcsp_probs[i]<<" ";
	cerr<<endl;
	vector<int> order;
	//Read evidence
	ifstream in(argv[2]);
	int num_evidence=0;
	vector<int> evidence;
	if(in.good())
	{
		in>>num_evidence;
		for(int i=0;i<num_evidence;i++)
		{
			int var,val;
			in>>var>>val;
			gm.variables[var]->value()=val;
			evidence.push_back(var);
		}
	}
	vector<set<int> > clusters;
	double estimate;
	double limit=(double)LIMIT;
	//sscanf(argv[4],"%lf",&limit);
	//sscanf(argv[5],"%d",&i_bound);
	bool run_samplesearch=false;
	gm.removeIrrelevantNetwork(evidence);
	if(gm.mode == DET)
	{
		gm.reduceDomains();
	}
	computeIBound(gm,i_bound);
	if(gm.mode == DET)
	{
		GM csp;
		gm.convertToCN(csp);
		Graph g;
		g.makeGraph(&csp);
		int treewidth = g.getMinFillOrdering(order);
		cerr<<"I-bound = "<<i_bound<<" Treewidth = "<<treewidth<<endl;
		//if(treewidth = i_bound)
			run_samplesearch=true;
			//else
			//gm.getClusters(order,clusters);
	}
	if(run_samplesearch || gm.mode == POSITIVE)
	{
		order=vector<int>();
		clusters=vector<set<int> > ();
		estimate=0.0;
		gm.getMinFillOrdering(order,clusters,estimate);
	}
	
	cerr<<"num evidence = "<<num_evidence<<endl;
	cout <<"t "<<get_cpu_time()<<endl;
	//Graph graph;
	//graph.makeGraph(&gm);
	//graph.getMinFillOrdering(order);
	
	//
	//BE be(gm.variables,gm.functions,order);

	// WCSP step
	// Find min amomg wcsp_probs
	Double min;
	bool first=true;
	for(int i=0;i<gm.wcsp_probs.size();i++){
	  if(first){
	    min=gm.wcsp_probs[i];
	    first=false;
	  }
	  else{
	    if(gm.wcsp_probs[i] < min)
	      min=gm.wcsp_probs[i];
	  }
	}


	if(is_wcsp && min < Double(0.1)){
	  cerr<<"IS WCSP\n";

	  // Create other GM
	  GM other_gm;
	  other_gm.variables=vector<Variable*>(gm.variables.size());
	  other_gm.type=gm.type;
	  other_gm.mode=gm.mode;
	  for(int i=0;i<gm.variables.size();i++){
	    vector<int> domain = gm.variables[i]->domain();
	    other_gm.variables[i]=new Variable(i,domain);
	  }
	  other_gm.functions=vector<Function*> (gm.functions.size());
	  
	  
	  
	    
	  for(int i=0;i<other_gm.functions.size();i++){
	    other_gm.functions[i]=new Function();
	    for(int j=0;j<gm.functions[i]->variables().size();j++){
	      int id=gm.functions[i]->variables()[j]->id();
	      other_gm.functions[i]->variables().push_back(other_gm.variables[id]);
	    }
	    other_gm.functions[i]->table()=gm.functions[i]->table();
	    	    for(int j=0;j<other_gm.functions[i]->table().size();j++){
	      if(other_gm.functions[i]->table()[j] > min){
		continue;
	      }
	      else {
		other_gm.functions[i]->table()[j]=Double();
		other_gm.mode=DET;
	      }
	    }
	  }
	  
	  cerr<<"Checking ---- \n";
	  // If the other GM is consistent then compute probability of evidence for this gm
	  if(SampleSearch::isconsistent(other_gm) && other_gm.mode==DET){
	    cerr<<"WCSP mode\n";
	    other_gm.reduceDomains();
	    computeIBound(other_gm,i_bound);
	    vector<int> other_order;
	    vector<set<int> > other_clusters;
	    double other_estimate=0.0;
	    vector<int> other_new_order;
	    other_gm.getMinFillOrdering(other_order,other_clusters,other_estimate);
	    other_gm.rearrangeOrdering(other_order,other_clusters,other_new_order,limit);
	    int other_i_bound = 2;
	    //computeIBound(other_gm,other_i_bound);
	    JG jg(other_gm,other_i_bound,num_iterations,other_order,LSS);

	    cerr<<"Join graph constructed\n";
	    jg.propagate();
	    RBSampleSearch ss;
	    Double pe1=ss.computePE(other_gm,jg,p_bound,other_order,other_new_order,num_samples);
	    pe1*=gm.mult_factor;
	    cerr<<"---------------------------RESULT------------------------------\n";
	    cerr<<"\t Log Base e probability of evidence = "<<log(pe1.value())<<endl;
	    cerr<<"\t Log Base 10 probability of evidence = "<<log(pe1.value())/log((double)10.0)<<endl;
	    cerr<<"\t Probability of evidence = "<<pe1<<endl;
	    cerr<<"---------------------------RESULT------------------------------\n";
	    return 1;
	
	  }
	}
	
	//for(int i=0;i<order.size();i++)
	//	cerr<<order[i]<<" ";
	//cerr<<endl;
	
	cerr<<"Limit = "<<limit<<endl;
	vector<int> new_order;
	gm.rearrangeOrdering(order,clusters,new_order,limit);
	cerr<<"#sampled variables = "<<new_order.size()<<" out of "<<gm.variables.size()<<endl;
	//BE be(gm.variables,gm.functions,order);

	//be.log_pe+=LogDouble(gm.mult_factor);
	//cerr<<"Estimate = "<<estimate<<endl;
	//cout<<"LogPe = "<<be.log_pe.val<<endl;


	//cout<<order.size()<<endl;
	JG jg(gm,i_bound,num_iterations,order,LSS);

	cerr<<"Join graph constructed\n";
	
	
	jg.propagate();
	

	if(!run_samplesearch)
	{
		RB_IJGP_Sampler sampler;
		Double pe=sampler.computePE(gm,jg,p_bound,order,new_order,time_limit,num_samples);
		pe*=gm.mult_factor;

		cerr<<"---------------------------RESULT------------------------------\n";
		cerr<<"\t Log Base e probability of evidence = "<<log(pe.value())<<endl;
		cerr<<"\t Log Base 10 probability of evidence = "<<log(pe.value())/log((double)10.0)<<endl;
		cerr<<"\t Probability of evidence = "<<pe<<endl;
		cerr<<"---------------------------RESULT------------------------------\n";
	}
	else
	{
		cerr<<"Running SampleSarch\n";
		RBSampleSearch ss;
		Double pe1=ss.computePE(gm,jg,p_bound,order,new_order,num_samples);
		pe1*=gm.mult_factor;
		cerr<<"---------------------------RESULT------------------------------\n";
		cerr<<"\t Log Base e probability of evidence = "<<log(pe1.value())<<endl;
		cerr<<"\t Log Base 10 probability of evidence = "<<log(pe1.value())/log((double)10.0)<<endl;
		cerr<<"\t Probability of evidence = "<<pe1<<endl;
		cerr<<"---------------------------RESULT------------------------------\n";
	}
}
