#include "GM.h"
#include "JG.h"
#include "IJGP_Sampler.h"
#include "SampleSearch.h"
#include "Graph.h"
#include <cmath>
#include <cstdlib>
#include <getopt.h>

using namespace std;

typedef enum
{
	MINFILL,TOPOLOGICAL,MINDEGREE
}ORDERING;



static struct option long_options[]=
{
	{"ordering",required_argument,NULL,'o'
	},
	{"i-bound",required_argument,NULL,'i'},
	{"task",required_argument,NULL,'w'},
	{"num-samples",required_argument,NULL,'n'},
	{"time-limit",required_argument,NULL,'t'},
	{"num-iterations",required_argument,NULL,'j'},
	{"input",required_argument,NULL,'f'},
	{"outfile",required_argument,NULL,'g'},
	{"help", no_argument,NULL,'h'},
};

static const char OPTION_STRING[]="o:i:w:n:t:j:f:hg:";

void print_help()
{
	cout<<"ijgpss [options] -f inputfile  \n";
	cout<<" --ordering <minfill,topological,mindegree> : What ordering to use\n";
	cout<<" --i-bound <int> : An integer value for the i-bound\n";
	cout<<" --num-iterations<int> : The maximum number of iterations for which IJGP is run\n";
	cout<<" --task <int>    : (0) for probability of evidence and (1) for belief updating\n";
	cout<<" --num-samples <int> : The maximum number of samples to be drawn\n";
	cout<<" --time-limt <int> : The maximum time in seconds for which the algorithm is run\n";
	cout<<" --outfile<string> : path where the output file is stored (default out) \n";
	cout<<" --help : Print this help\n";
}

int main(int argc, char *argv[])
{

	char ordering[100];
	int i_bound=3,p_bound=0,w_bound=0,num_samples=-1,time_limit=10,num_iterations=10;
	GM gm;
	char graphfile[100];
	bool writegraph=false;
	bool p_of_e=true;
	int task;
	char outfilename[100]="out";
	if(argc==1)
	{
		print_help();
		return 1;
	}
	while(1)
	{

		int option_index =0;
		int opt = getopt_long(argc,argv,OPTION_STRING,long_options,&option_index);
		// Detect end of options
		if (opt == -1)
			break;
		switch(opt)
		{
		case 'o':
			sscanf(optarg,"%s",&ordering);
			break;
		case 'i':
			sscanf(optarg,"%ld",&i_bound);
			break;

		case 'n':
			sscanf(optarg,"%ld",&num_samples);
			break;
		case 't':
			sscanf(optarg,"%ld",&time_limit);
			break;
		case 'w':
			sscanf(optarg,"%d",&task);
			if(task > 0)
				p_of_e=false;
			else
				p_of_e=true;
			break;
		case 'j':
			sscanf(optarg,"%ld",&num_iterations);
			break;

		case 'f':
			if(!gm.readCNF(optarg))
			{
				gm=GM();
				//gm.readVIB(optarg);
				gm.readErgo(optarg);
				char evidfile[100];
				sprintf(evidfile,"%s.evid",optarg);
				//gm.readEvidence(evidfile);
				gm.readEvidenceErgo(evidfile);
			}
			break;
		case 'g':
			strcpy(outfilename,optarg);
			break;
		case 'h':
			print_help();
			return 1;
		default:
			print_help();
			return 1;
		}
	}
	cerr<<"i-bound= "<<i_bound<<endl;

	vector<int> order(gm.variables.size());
	if(strcmp(ordering,"minfill")==0)
	{
		Graph g;
		g.makeGraph(&gm);
		g.getMinFillOrdering(order);
	}
	else if(strcmp(ordering,"topological")==0)
	{
		vector<int> sampling_order (order.size());
		//Graph g;
		//g.makeDirectedGraph(&gm);
		//g.getTopologicalOrdering(sampling_order);
		for(int i=0;i<sampling_order.size();i++)
		{
			sampling_order[i]=i;
		}
		for(int i=0;i<sampling_order.size();i++)
		{
			order[i]=sampling_order[order.size()-i-1];
		}
	}
	else //if (strcmp(ordering,"rcordering")==0)
	{
		GM csp;
		gm.convertToCN(csp);
		Graph g;
		g.makeGraph(&csp);
		g.getMinFillOrdering(order);
	}
	
	JG jg(gm,i_bound,num_iterations,order,SS);
	cerr<<"Join graph constructed\n";
	time_t start_time,end_time;
	start_time=time(NULL);
	jg.propagate();
	end_time=time(NULL);

	if(p_of_e)
	{
		SampleSearch ss(gm,jg,0,order);
		Double pe=ss.computePE(time_limit,num_samples);
		pe*=gm.mult_factor;
		ofstream outfile(outfilename);
		outfile<<log(pe.value())<<endl;
		outfile.close();
		cout<<"---------------------------RESULT------------------------------\n";
		cout<<"\t Log probability of evidence = "<<log(pe.value())<<endl;
		cout<<"\t Probability of evidence = "<<pe<<endl;
		cout<<"---------------------------RESULT------------------------------\n";

	}
	else
	{
		vector<vector<Double> > beliefs;
		SampleSearch ss(gm,jg,0,order);
		ss.computeBeliefs(time_limit,num_samples,beliefs);
		ofstream outfile(outfilename);
		outfile<<beliefs.size()<<endl;
		for(int i=0;i<beliefs.size();i++)
		{
			outfile<<i<<" "<<beliefs[i].size()<<" ";
			for(int j=0;j<beliefs[i].size();j++)
			{
				outfile<<beliefs[i][j]<<" ";
			}
			outfile<<endl;
		}
		outfile.close();
	}

}