#include <iostream>
#include <fstream>
#include "SampleSearch.cpp"
#include "BE.h"
#include "GM.h"
#include "Graph.h"
#include "RB_IJGP_Sampler.h"
#include <cmath>
#include <string>

using namespace std;
#include <sys/time.h>
#include <sys/resource.h>
static inline double get_cpu_time(void) {
	struct rusage ru;
	getrusage(RUSAGE_SELF, &ru);
	return (double) ru.ru_utime.tv_sec + (double) ru.ru_utime.tv_usec / 1000000;
}

int LIMIT = 10000000;

void spec_order_to_order(GM& gm, vector<int>& spec_order, vector<int>& order) {
	map<int, int> orig2curr;
	for (int i = 0; i < gm.variables.size(); i++) {
		orig2curr[gm.variables[i]->orig_id] = i;
	}
	order = vector<int> ();
	for (int i = 0; i < spec_order.size(); i++) {
		map<int, int>::iterator iter = orig2curr.find(spec_order[i]);
		if (iter != orig2curr.end()) {
			order.push_back(iter->second);
		}
	}
}
int main(int argc, char* argv[]) {
	int i_bound, p_bound = 0, w_bound = 0, num_samples = -1, time_limit = 10,
			num_iterations = 10;
	GM gm;
	if (argc < 3) {
		cerr
				<< "Format is vec <uai-file> <uai-evid-file> <random-seed> <optional ordering file> <optional memory-bound> \n";
		cerr
				<< " memory bound specifies the maximum amount of memory that can be used by each bucket in Bucket (Variable) Elimination.\n";
		cerr
				<< "Default value of memory bound is 10000000, namely 10 Million, increase it if you have more RAM\n";
		exit(-1);
	}
	//gm.readErgo(argv[1]);
	gm.readUAI08(argv[1]);
	vector<int> order;
	//Read evidence
	ifstream in(argv[2]);
	//gm.mode=POSITIVE;
	//char buffer[100];
	//in.getline(buffer,100);
	int num_evidence = 0;
	vector<int> evidence;
	if (in.good()) {
		in >> num_evidence;
		for (int i = 0; i < num_evidence; i++) {
			int var, val;
			in >> var >> val;
			gm.variables[var]->value() = val;
			evidence.push_back(var);
		}
	}
	vector<set<int> > clusters;
	double estimate;
	double limit = (double) LIMIT;
	vector<int> spec_order;
	bool is_order_given = false;
	if (argc > 4) {
		is_order_given = true;
		ifstream ord_file(argv[4]);
		int ord_num_vars, ord_treewidth;
		string line;
		// Ignore the first line of the file
		getline(ord_file,line);
		
		ord_file >> ord_num_vars;
		ord_file >> ord_treewidth;
		if ((int) gm.variables.size() != ord_num_vars) {
			cerr
					<< "Num variables in ordering file do not match the num variables in the GM\n";
			cerr
					<< "This is probably due to the singleton consistency checks in VEC\n";
			cerr
					<< "Do not specify the ordering file for this instance and then run VEC\n";
			exit(-1);
		}
		spec_order = vector<int> (ord_num_vars);
		for (int i = ord_num_vars - 1; i > -1; i--)
			ord_file >> spec_order[i];
		ord_file.close();
	}
	if (argc > 5) {
		limit = atoi(argv[5]);
	}
	//sscanf(argv[4],"%lf",&limit);
	//sscanf(argv[5],"%d",&i_bound);
	if (gm.type == BAYES) {
		gm.removeIrrelevantNetwork(evidence);

		if (gm.mode == DET) {
			gm.reduceDomains();
		}
		//	GM csp;
		//	gm.convertToCN(csp);
		//	Graph g;
		//	g.makeGraph(&csp);
		//	g.getMinFillOrdering(order);
		//	gm.getClusters(order,clusters);
		//}
		//else
		//{
		spec_order_to_order(gm, spec_order, order);
		if (!is_order_given) {
			gm.getMinFillOrdering(order, clusters, estimate);
		} else {
			gm.getClusters(order, clusters);
		}
		//	}
	} else if (gm.type == MARKOV) {
		gm.removeIrrelevantNetwork(evidence);
		if (gm.mode == DET) {
			gm.reduceDomains();
		}
		spec_order_to_order(gm, spec_order, order);
		if (!is_order_given) {
			gm.getMinFillOrdering(order, clusters, estimate);
		} else {
			gm.getClusters(order, clusters);
		}

	}
	cerr << "num evidence = " << num_evidence << endl;
	//Graph graph;
	//graph.makeGraph(&gm);
	//graph.getMinFillOrdering(order);

	//
	//BE be(gm.variables,gm.functions,order);


	cerr << "Limit = " << limit << endl;
	vector<int> new_order;
	gm.rearrangeOrdering(order, clusters, new_order, limit);
	cerr << "#sampled variables = " << new_order.size() << " out of "
			<< gm.variables.size() << endl;

	//computeIBound(gm,i_bound);
	vector<Variable*> cond_variables(new_order.size());
	for (int i = 0; i < new_order.size(); i++)
		cond_variables[i] = gm.variables[new_order[i]];
	int num_values = Variable::getDomainSize(cond_variables);
	//Double pe;
	long double pe = 0.0;
	cerr << "num values = " << num_values << endl;
	Solver solver;

	if (gm.mode == DET) {
		gm.convertToSAT();
		gzFile in = gzopen(satfile, "rb");
		parse_DIMACS(in, solver);
		gzclose(in);
	}
	int n = cond_variables.size();
	int a[n];
	int f[n + 1];
	int o[n];
	int m[n];
	// Initialize
	for (int i = 0; i < n; i++) {
		cond_variables[i]->value() = 0;
		a[i] = 0;
		f[i] = i;
		o[i] = 1;
		m[i] = cond_variables[i]->domain_size();
	}
	f[n] = n;
	cout << "t " << get_cpu_time() << endl;

	while (1) {
		bool has_sol = true;
		if (gm.mode == DET) {
			vec<Lit> assumptions;
			for (int j = 0; j < cond_variables.size(); j++) {
				int id = cond_variables[j]->id();
				int value = cond_variables[j]->value();
				assumptions.push(Lit(gm.csp_to_sat_variables[id][value]));
			}
			if (!solver.solve(assumptions)) {
				has_sol = false;
			}
		}
		if (has_sol) {
			BE be(gm.variables, gm.functions, order);
			//pe+=be.log_pe.toDouble();
			pe += exp(be.log_pe.toLongdouble());
		}

		//Double tmp(pe);
		//tmp*=gm.mult_factor;
		long double tmp = pe;
		tmp *= gm.mult_factor.value();
		if (tmp > 0.0)
			cout << "z " << log10(tmp) << endl;
		int j = f[0];
		f[0] = 0;
		if (j == n)
			break;
		a[j] = a[j] + o[j];
		cond_variables[j]->value() = a[j];
		if (a[j] == 0 || a[j] == (m[j] - 1)) {
			o[j] = -o[j];
			f[j] = f[j + 1];
			f[j + 1] = j + 1;
		}
	}

	//pe*=gm.mult_factor;
	pe *= gm.mult_factor.value();
	cerr << "---------------------------RESULT------------------------------\n";
	cerr << "\t Log Base e probability of evidence = " << log(pe) << endl;
	cerr << "\t Log Base 10 probability of evidence = " << log(pe) / log(
			(double) 10.0) << endl;
	cerr << "\t Probability of evidence = " << pe << endl;
	cerr << "---------------------------RESULT------------------------------\n";

}
