#include "ClauseList.h"
#include "Clause.h"
#include "SATInstance.h"
#include "BigNum.h"
//#include "SATSolver.h"
#include "SampleSearch.h"
#include "Solver.h"
#include "BE.h"
#include "JT.h"
#include <zlib.h>
#include "ORStore.h"


char satfile[100]="temp.cnf";
string printbytes(int bytes)
{
	char tmp[100];
	double gb=-5.0;
	double mb;
	if (bytes > 1073741824){
		gb=(double)bytes/(double)1073741824;	
	}
	
	mb=(double)bytes/(double)1048576;
	
	if (gb > 0){
		sprintf(tmp,"%fGB",gb);
		
	}
	else
	{
		sprintf(tmp,"%fMB",mb);
	}
	string tmp1=tmp;
	return tmp1;
}
double checkNumSolutions(GM& gm, vec<Lit>& assumptions)
{
	ofstream out("sat_tmp.cnf");
	out<<"p cnf "<<gm.variables.size()<<" "<<gm.clauses.size()+assumptions.size()<<endl;
	for(int i=0;i<gm.clauses.size();i++){
		for(int j=0;j<gm.clauses[i].size();j++){
			int Var=var(gm.clauses[i][j]);
			int p=(Lit(Var)==gm.clauses[i][j])?(Var+1):(-(Var+1));
			out<<p<<" ";
		}
		out<<0<<endl;
	}
	for(int i=0;i<assumptions.size();i++){
		int Var=var(assumptions[i]);
		int p=(Lit(Var)==assumptions[i])?(Var+1):(-(Var+1));
		out<<p<<" 0\n";
	}
	out.close();
	system("/home/vibhav/work/cachet/cachet sat_tmp.cnf -q > t");
	ifstream in("t");
	double k;
	in>>k;
	//system(" cat t");
	in.close();
	if (k <=0.0){
		exit(1);
	}
	return k;
	
}
bool normalize_table(vector<Double>& table)
{
	Double norm_const;
	for(int i=0;i<table.size();i++){
		norm_const+=table[i];
	}
	for(int i=0;i<table.size();i++){
		table[i]/=norm_const;
	}		
}
int generateSample(vector<Double>& table, myRandom& random, bool normalize=true)
{
	if (normalize){
		normalize_table(table);		
	}
	int value=INVALID_VALUE;
	double rand_num=random.getDouble();
	double cdf=0.0;
	double curr_var_weight=0.0;
	for(int k=0;k<table.size();k++)
	{
		if(!table[k].isZero())
		{
			value=k;
			cdf+=table[k].value();
			curr_var_weight=table[k].value();
			if(rand_num <= cdf)
			{
				value=k;
				break;
			}
		}
	}
	return value;
}
void updateInferredAssignmentsSAT(GM& gm, vec<Lit>& assumptions, set<Lit>& inferred_assignments, Solver& solver)
{
	for(int i=0;i<assumptions.size();i++){
		solver.inferred_assignments.erase(assumptions[i]);
	}
	for(set<Lit>::iterator i=solver.inferred_assignments.begin();i!=solver.inferred_assignments.end();i++){
		inferred_assignments.insert(*i);
	}
	for(set<Lit>::iterator i=inferred_assignments.begin();i!=inferred_assignments.end();i++){
		int Var=var(*i);
		Lit q= Lit(Var);
		int value=(*i==q)?(1):(0);
		assumptions.push(*i);
		if(gm.variables[Var]->value()==INVALID_VALUE){
			gm.variables[Var]->value()=value;
		}
		else{
			assert(gm.variables[Var]->value()==value);
		}
	}
}
bool isConsistent(int var, int value, vec<Lit>& assumptions,ORStore& store, Solver& solver)
{
	if(store.check(var,value,false)==INCONSISTENT){
		return false;
	}
	if(store.check(var,value,false)==CONSISTENT){
		store.AddAssignment(var,value);
		return true;
	}
	assert(store.check(var,value,false)==DONTKNOW);
	if(solver.solve(assumptions)){
		store.AddAssignment(var,value);
		//updateInferredAssignmentsSAT(gm,assumptions,inferred_assignments,solver);
		return true;
	}
	store.AddNoGood(var,value);
	//updateInferredAssignmentsSAT(gm,assumptions,inferred_assignments,solver);
	return false;
}
bool isConsistentSAT(GM& gm, int var, int value, vec<Lit>& assumptions,set<Lit>& inferred_assignments,ORStore& store, Solver& solver)
{
	int other_value=(value==0)?(1):(0);
	if(store.check(var,value,false)==INCONSISTENT){	
		assert(store.check(var,other_value,false)==CONSISTENT);
		return false;
	}
	if(store.check(var,value,false)==CONSISTENT){
		store.AddAssignment(var,value);
		return true;
	}
	assert(store.check(var,value,false)==DONTKNOW);
	if(solver.solve(assumptions)){
		store.AddAssignment(var,value);
		//updateInferredAssignmentsSAT(gm,assumptions,inferred_assignments,solver);
		return true;
	}
	store.AddNoGood(var,value);
	store.AddGood(var,other_value);
	//updateInferredAssignmentsSAT(gm,assumptions,inferred_assignments,solver);
	return false;
}
//=================================================================================================
// DIMACS Parser:

#define CHUNK_LIMIT 1048576

class StreamBuffer {
    gzFile  in;
    char    buf[CHUNK_LIMIT];
    int     pos;
    int     size;

    void assureLookahead() {
        if (pos >= size) {
            pos  = 0;
            size = gzread(in, buf, sizeof(buf)); } }

public:
    StreamBuffer(gzFile i) : in(i), pos(0), size(0) {
        assureLookahead(); }

    int  operator *  () { return (pos >= size) ? EOF : buf[pos]; }
    void operator ++ () { pos++; assureLookahead(); }
};

//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template<class B>
static void skipWhitespace(B& in) {
    while ((*in >= 9 && *in <= 13) || *in == 32)
        ++in; }

template<class B>
static void skipLine(B& in) {
    for (;;){
        if (*in == EOF || *in == '\0') return;
        if (*in == '\n') { ++in; return; }
        ++in; } }

template<class B>
static int parseInt(B& in) {
    int     val = 0;
    bool    neg = false;
    skipWhitespace(in);
    if      (*in == '-') neg = true, ++in;
    else if (*in == '+') ++in;
    if (*in < '0' || *in > '9') reportf("PARSE ERROR! Unexpected char: %c\n", *in), exit(3);
    while (*in >= '0' && *in <= '9')
        val = val*10 + (*in - '0'),
        ++in;
    return neg ? -val : val; }

template<class B>
static void readClause(B& in, Solver& S, vec<Lit>& lits) {
    int     parsed_lit, var;
    lits.clear();
    for (;;){
        parsed_lit = parseInt(in);
        if (parsed_lit == 0) break;
        var = abs(parsed_lit)-1;
        while (var >= S.nVars()) S.newVar();
        lits.push( (parsed_lit > 0) ? Lit(var) : ~Lit(var) );
    }
}

template<class B>
static bool match(B& in, char* str) {
    for (; *str != 0; ++str, ++in)
        if (*str != *in)
            return false;
    return true;
}


template<class B>
static void parse_DIMACS_main(B& in, Solver& S) 
{
    vec<Lit> lits;
    for (;;){
        skipWhitespace(in);
        if (*in == EOF)
            break;
        else if (*in == 'p'){
            if (match(in, "p cnf")){
                int vars    = parseInt(in);
                int clauses = parseInt(in);
                //reportf("|  Number of variables:  %-12d                                         |\n", vars);
                //reportf("|  Number of clauses:    %-12d                                         |\n", clauses);
            }else{
                reportf("PARSE ERROR! Unexpected char: %c\n", *in), exit(3);
            }
        } else if (*in == 'c' || *in == 'p')
            skipLine(in);
        else
            readClause(in, S, lits),
            S.addClause(lits);
    }
}

// Inserts problem into solver.
//
static void parse_DIMACS(gzFile input_stream, Solver& S) 
{
    StreamBuffer in(input_stream);
    parse_DIMACS_main(in, S); 
}


//=================================================================================================


SampleSearch::SampleSearch(GM& gm_, JG& jg_, int p_bound, vector<int>& order_): 
gm(gm_),jg(jg_)
{
	helper=IJGPSamplingHelper(&gm,&jg,0,order_);
	order=vector<int> (order_.size());
	for(int i=0;i<order.size();i++)
		order[order.size()-i-1]=order_[i];
	gm.convertToSAT();
	
	/*if(strcmp(satfile,"temp.cnf")==0)
	{
		SATInstance xInstance;
		SATSolver xSATSolver(&xInstance);
		xInstance.bReadDimacs("temp.cnf");
		xSATSolver.bPreprocess(3,10);
		ofstream xNewInstance;
		xNewInstance.open("temp.cnf", ios::out);
		xInstance.vOutputDimacs(xNewInstance);
	}*/
	
}

void SampleSearch::computeBeliefs(int time_limit, int num_samples, vector<vector<Double> >& beliefs)
{
	Double weight;
	myRandom random;
	time_t start_time,curr_time;
	start_time=time(NULL);
	Solver solver;
	gzFile in = gzopen(satfile, "rb");
	parse_DIMACS(in,solver);
	gzclose(in);	

	beliefs=vector<vector<Double> > (gm.variables.size());
	for(int i=0;i<gm.variables.size();i++)
	{
		beliefs[i]=vector<Double> (gm.variables[i]->domain_size());
	}
	
	for(int samples=0;samples!=num_samples;samples++)
	{
		
		solver.sampler=this;
		curr_time=time(NULL);
		if(num_samples==INVALID_VALUE && (curr_time-start_time) >= time_limit)
		{
			break;
		}
		// Initialize
		for(int i=0;i<gm.variables.size();i++)
		{
			gm.variables[i]->value()=INVALID_VALUE;
		}
		bool ret_value=solver.solve_sample();
		assert(ret_value);
		cerr<<"Sample "<<samples<<" generated \n";
		Double curr_weight=getWeight(solver.model);
		for(int i=0;i<gm.variables.size();i++)
		{
			assert(gm.variables[i]->value()!=INVALID_VALUE);
			beliefs[i][gm.variables[i]->value()]+=curr_weight;	
		}	
		weight+=curr_weight;		
		cerr<<"\tSample "<<samples<<" weight computed\n";
	}

	for(int i=0;i<gm.variables.size();i++)
	{
		gm.variables[i]->value()=INVALID_VALUE;
		for(int j=0;j<beliefs[i].size();j++)
			beliefs[i][j]/=weight;
	}
	
}

Double SampleSearch::computePE(int time_limit, int num_samples)
{
	Double weight;
	myRandom random;
	time_t start_time,curr_time;
	start_time=time(NULL);
	Solver solver;
	gzFile in = gzopen(satfile, "rb");
	parse_DIMACS(in,solver);
	gzclose(in);
	
	
	
	for(int samples=0;samples!=num_samples;samples++)
	{
		
		solver.sampler=this;
		curr_time=time(NULL);
		if(num_samples==INVALID_VALUE && (curr_time-start_time) >= time_limit)
		{
			return weight/Double((double)samples);
		}
		// Initialize
		for(int i=0;i<gm.variables.size();i++)
		{
			gm.variables[i]->value()=INVALID_VALUE;
		}
		bool ret_value=solver.solve_sample();
		assert(ret_value);
		cerr<<"Sample "<<samples<<" generated \n";
		Double curr_weight=getWeight(solver.model);
		for(int i=0;i<gm.variables.size();i++)
		{
			assert(gm.variables[i]->value()!=INVALID_VALUE);
			
		}	
		
		weight+=curr_weight;
		cerr<<"\tSample "<<samples<<" weight computed\n";
	}
	return weight/Double((double)num_samples);
}
void SampleSearch::getSample(vector<Double>& bf)
{
	// The following should be changed....very bad hack
	
	
	
	Solver solver;
	gzFile in = gzopen(satfile, "rb");
	parse_DIMACS(in,solver);
	gzclose(in);
	// End...change this

	solver.sampler=this;
	bool ret_value=solver.solve_sample();
	assert(ret_value);
	getWeight(solver.model,bf);
}
int SampleSearch::getSample(int& variable, bool& value,vec<char>& assignment)
{
	
	int csp_variable=INVALID_VALUE;
	// Select the earliest unassigned variable

	for(int i=0;i<gm.variables.size();i++)
	{
		int var=order[i];
		
		for(int j=0;j<gm.csp_to_sat_variables[var].size();j++)
		{
			if(toLbool(assignment[gm.csp_to_sat_variables[var][j]])==l_Undef)
			{
				csp_variable=var;
				//cerr<<var<<endl;
				break;
			}
		}
		if(csp_variable==var)
			break;
		
	}
	
	// If all varables are assigned a value return
	if(csp_variable==INVALID_VALUE)
	{
		variable=-1;
		value=false;
		return 1;
	}
	gm.variables[csp_variable]->value()=INVALID_VALUE;
	Double cdf,norm_const;
	vector<Double> func_table(gm.variables[csp_variable]->domain_size());
	//Set values of all variables in the function
	// Assign values to all variables of the function
	
	//SF* function=&helper.sampling_functions[csp_variable];
	Function* function=&helper.getFunction(csp_variable);
	
	for(int i=0;i<function->variables().size();i++)
	{
		if(function->variables()[i]->id()==csp_variable)
			continue;
		function->variables()[i]->value()=INVALID_VALUE;
		int var=function->variables()[i]->id();
		bool val_set=false;
		//function->[i]->resetCurrDomain();
		
		for(int j=0;j<gm.csp_to_sat_variables[var].size();j++)
		{
			if(toLbool(assignment[gm.csp_to_sat_variables[var][j]])==l_True)
			{
				val_set=true;
				gm.variables[var]->value()=j;
				gm.variables[var]->addr_value()=j;
				break;
			}
		}
		assert(val_set);
	}
	assert((int)gm.csp_to_sat_variables[csp_variable].size()==gm.variables[csp_variable]->domain_size());
	int default_value=INVALID_VALUE;
	for(int i=0;i<gm.variables[csp_variable]->domain_size();i++)
	{
		if(toLbool(assignment[gm.csp_to_sat_variables[csp_variable][i]])==l_False)
		{
			continue;
		}
		default_value=i;
		gm.variables[csp_variable]->addr_value()=i;
		int address=Variable::getAddress(function->variables());
		func_table[i]=function->table()[address];
		norm_const+=func_table[i];
	}
	assert(default_value!=INVALID_VALUE);
	for(int i=0;i<gm.variables[csp_variable]->domain_size();i++)
	{
		func_table[i]/=norm_const;
	}
	double rand_val=random.getDouble();
	int sampled_value=INVALID_VALUE;
	for(int i=0;i<gm.variables[csp_variable]->domain_size();i++)
	{
		cdf+=func_table[i];
		if(cdf.value() >=rand_val)
		{
			sampled_value=i;
			//cerr<<func_table[i]<<" ";
			break;
		}
	}
	if(sampled_value==INVALID_VALUE)
	{
		sampled_value=default_value;
	}

	assert(sampled_value!=INVALID_VALUE);
	variable=gm.csp_to_sat_variables[csp_variable][sampled_value];
	value=false;
	
}

void SampleSearch::getWeight(vec<lbool>& assignment,vector<Double>& bf)
{
	//Double sample_weight(1.0),actual_weight(1.0);
	Solver solver;
	gzFile in = gzopen(satfile, "rb");
	parse_DIMACS(in,solver);
	gzclose(in);
	vec<Lit> assumptions;
	bf=vector<Double> (gm.variables.size());
	//cout<<assignment.size()<<endl;
	// Set value
	for(int v=0;v<gm.variables.size();v++)
	{
		int var=order[v];
		//cout<<v<<" "<<flush;
		vector<Double> func_table(gm.variables[var]->domain_size());
		
		Function& function=helper.getFunction(var);
		//SF& function=
		for(int i=0;i<gm.csp_to_sat_variables[var].size();i++)
		{
			gm.variables[var]->addr_value()=i;
			int address=Variable::getAddress(function.variables());
			func_table[i]=function.table()[address];
		}
		bool val_set=false;
		
		for(int i=0;i<gm.csp_to_sat_variables[var].size();i++)
		{
			//cout<<gm.csp_to_sat_variables[var][i]<<" "<<flush;
		
			if(assignment[gm.csp_to_sat_variables[var][i]]==l_True)
			{
				val_set=true;
				gm.variables[var]->addr_value()=i;
				gm.variables[var]->value()=i;
				
				//continue;
			}
			else
			{
				assumptions.push((Lit(gm.csp_to_sat_variables[var][i])));
				//gzFile in = gzopen(satfile, "rb");
				//parse_DIMACS(in,solver);
				//cout<<"checking --- "<<flush;
				if(!solver.solve(assumptions))
				{
					func_table[i]=Double();
				}
				//cout<<"Done\n";
				assumptions.pop();
			}
		}
		assert(val_set);
		Double norm_const;
		// Normalize the fun_table
		for(int i=0;i<func_table.size();i++)
		{
			norm_const+=func_table[i];
			
		}
		bf[var]=(func_table[gm.variables[var]->addr_value()]/norm_const);
		//sample_weight*=(func_table[gm.variables[var]->addr_value()]/norm_const);
		
		//cout<<sample_weight<<" "<<func_table[gm.variables[var]->addr_value()]/norm_const<<endl;
		assumptions.push(Lit(gm.csp_to_sat_variables[var][gm.variables[var]->addr_value()]));
	}

	//for(int i=0;i<gm.functions.size();i++)
	//{
	//	actual_weight*=gm.functions[i]->getWeight();
	//}
	////cout<<"------"<<sample_weight<<" "<<actual_weight<<endl;
	//return actual_weight/sample_weight;
}

Double SampleSearch::getWeight(vec<lbool>& assignment)
{
	Double sample_weight(1.0),actual_weight(1.0);
	Solver solver;
	gzFile in = gzopen(satfile, "rb");
	parse_DIMACS(in,solver);
	gzclose(in);
	vec<Lit> assumptions;
	//cout<<assignment.size()<<endl;
	// Set value
	for(int v=0;v<gm.variables.size();v++)
	{
		int var=order[v];
		//cout<<v<<" "<<flush;
		vector<Double> func_table(gm.variables[var]->domain_size());
		
		Function& function=helper.getFunction(var);
		//SF& function=
		for(int i=0;i<gm.csp_to_sat_variables[var].size();i++)
		{
			gm.variables[var]->addr_value()=i;
			int address=Variable::getAddress(function.variables());
			func_table[i]=function.table()[address];
		}
		bool val_set=false;
		
		for(int i=0;i<gm.csp_to_sat_variables[var].size();i++)
		{
			//cout<<gm.csp_to_sat_variables[var][i]<<" "<<flush;
		
			if(assignment[gm.csp_to_sat_variables[var][i]]==l_True)
			{
				val_set=true;
				gm.variables[var]->addr_value()=i;
				gm.variables[var]->value()=i;
				
				//continue;
			}
			else
			{
				assumptions.push((Lit(gm.csp_to_sat_variables[var][i])));
				//gzFile in = gzopen(satfile, "rb");
				//parse_DIMACS(in,solver);
				//cout<<"checking --- "<<flush;
				if(!solver.solve(assumptions))
				{
					func_table[i]=Double();
				}
				//cout<<"Done\n";
				assumptions.pop();
			}
		}
		assert(val_set);
		Double norm_const;
		// Normalize the fun_table
		for(int i=0;i<func_table.size();i++)
		{
			norm_const+=func_table[i];
			
		}
		
		sample_weight*=(func_table[gm.variables[var]->addr_value()]/norm_const);
		
		//cout<<sample_weight<<" "<<func_table[gm.variables[var]->addr_value()]/norm_const<<endl;
		assumptions.push(Lit(gm.csp_to_sat_variables[var][gm.variables[var]->addr_value()]));
	}

	for(int i=0;i<gm.functions.size();i++)
	{
		actual_weight*=gm.functions[i]->getWeight();
	}
	//cout<<"------"<<sample_weight<<" "<<actual_weight<<endl;
	return actual_weight/sample_weight;
}


RBSampleSearch::RBSampleSearch()
{
}

void RBSampleSearch::reduce(GM &gm, std::vector<vector<bool> >& new_domains)
{
	new_domains=vector<vector<bool> > (gm.variables.size());
	for(int i=0;i<new_domains.size();i++)
		new_domains[i]=vector<bool>(gm.variables[i]->domain_size());

	gm.convertToSAT();

	Solver solver;
	gzFile in = gzopen(satfile, "rb");
	parse_DIMACS(in,solver);
	gzclose(in);
	
	cerr<<"Solver inited\n";
	for(int i=0;i<new_domains.size();i++)
	{
		for(int j=0;j<new_domains[i].size();j++)
		{
			if(new_domains[i][j])
				continue;
			vec<Lit> assumptions;
			assumptions.push((Lit(gm.csp_to_sat_variables[i][j])));
			if(solver.solve(assumptions))
			{
					new_domains[i][j]=true;
			}
			/*vec<lbool>& assignment=solver.model;
			for(int a=0;a<new_domains.size();a++)
				for(int b=0;b<new_domains[a].size();b++)
					if(assignment[gm.csp_to_sat_variables[a][b]]==l_True)
					{
						new_domains[a][b]=true;
					}*/
		}
	}
	cerr<<"Domains reduced\n";
}
Double RBSampleSearch::computePE(GM& gm, JG& jg, int p_bound, vector<int>& order,vector<int>& sampling_order, int num_samples)
{
	gm.convertToSAT();

	Solver solver;
	gzFile in = gzopen(satfile, "rb");
	parse_DIMACS(in,solver);
	gzclose(in);
	
	cerr<<"Solver inited\n";

	vector<bool> is_sampled(order.size());
	for(int i=0;i<sampling_order.size();i++)
		is_sampled[sampling_order[i]]=true;
	vector<int> helper_order;
	vector<int> helper_sampling_order;
	
	for(int i=0;i<order.size();i++)
		if(!is_sampled[order[i]])
			helper_order.push_back(order[i]);
		else
			helper_sampling_order.push_back(order[i]);
	//for(int i=(int)sampling_order.size()-1;i>-1;i--)
	//	helper_order.push_back(sampling_order[i]);
	for(int i=0;i<helper_sampling_order.size();i++)
		helper_order.push_back(helper_sampling_order[i]);
	sampling_order=vector<int>();
	for(int i=helper_sampling_order.size()-1;i>-1;i--)
		sampling_order.push_back(helper_sampling_order[i]);
	assert(helper_order.size()==order.size());
	IJGPSamplingHelper helper(&gm,&jg,p_bound,helper_order);
	myRandom random;
	Double weight;
	time_t curr_time;
	for(int i=0;i<num_samples;i++)
	{
		for(int j=0;j<sampling_order.size();j++)
			gm.variables[j]->value()=INVALID_VALUE;

		vec<Lit> assumptions;
		double sample_weight=0.0;
		for(int j=0;j<sampling_order.size();j++)
		{
			int var=sampling_order[j];
			vector<Double> func_table(gm.variables[var]->domain_size());
			Function& function=helper.getFunction(var);
			//if(j==0)
			//cerr<<"tab-size = "<<function.table().size()<<endl;
			for(int k=0;k<gm.csp_to_sat_variables[var].size();k++)
			{
				gm.variables[var]->addr_value()=k;
				int address=Variable::getAddress(function.variables());
				func_table[k]=function.table()[address];
				//cerr<<func_table[k]<<endl;
			}
			for(int k=0;k<gm.csp_to_sat_variables[var].size();k++)
			{
				assumptions.push((Lit(gm.csp_to_sat_variables[var][k])));
				if(!solver.solve(assumptions))
				{
					func_table[k]=Double();
				}
				//cout<<"Done\n";
				assumptions.pop();
			}
			Double norm_const;
			// Normalize the fun_table
			for(int k=0;k<func_table.size();k++)
			{
				norm_const+=func_table[k];
			}
			for(int k=0;k<func_table.size();k++){
			  func_table[k]/=norm_const;
			}
			int value=INVALID_VALUE;
			double rand_num=random.getDouble();
			double cdf=0.0;
			double curr_var_weight=0.0;
			for(int k=0;k<func_table.size();k++)
			{
				if(!func_table[k].isZero())
				{
					value=k;
					cdf+=func_table[k].value();
					curr_var_weight=func_table[k].value();
					if(rand_num <= cdf)
					{
						value=k;
						break;
					}
				}
			}
			assert(value!=INVALID_VALUE);
			gm.variables[var]->value()=value;
			assumptions.push((Lit(gm.csp_to_sat_variables[var][value])));
			sample_weight+=log(curr_var_weight);
		}
		BE be(gm.variables,gm.functions,order);
		double actual_weight=log(be.log_pe.toDouble().value());
		weight+=(exp(actual_weight-sample_weight));
		Double pe=weight/Double((double)i+1);
		pe*=gm.mult_factor;
		cout.setf(ios::fixed,ios::floatfield);
		time_t some_time=time(NULL);
		if(curr_time != some_time){
		 curr_time=time(NULL);
		 cout<<" z "<<log10(pe.value())<<endl;
		}
		for(int j=0;j<gm.variables.size();j++){
		  gm.variables[j]->value()=INVALID_VALUE;
		}
	}
	for(int j=0;j<gm.variables.size();j++)
	{
		gm.variables[j]->value()=INVALID_VALUE;
	}
	return weight/Double((double)num_samples);
}

pair<double,double> purgeORStore(GM& gm, IJGPSamplingHelper& helper, ORStore& store, vector<int>& sampling_order, vector<vector<int> >& all_samples, vector<double>& all_weights)
{
    
	assert(all_samples.size()==all_weights.size());
	double overall_weight_ub=0.0;
	double overall_weight_lb=0.0;
	for(int i=0;i<all_samples.size();i++){
		double sample_weight_lb=0.0,sample_weight_ub=0.0;
		for(int j=0;j<sampling_order.size();j++){
		  gm.variables[sampling_order[j]]->value()=INVALID_VALUE;
		}
		assert(sampling_order.size() == all_samples[i].size());
		for(int j=0;j<all_samples[i].size();j++){
			int var=sampling_order[j];
			int val=all_samples[i][j];
			vector<Double> func_table_lb(gm.variables[var]->domain_size());
			vector<Double> func_table_ub(gm.variables[var]->domain_size());
			Function& function=helper.getFunction(var);
			for(int k=0;k<gm.csp_to_sat_variables[var].size();k++)
			{
				func_table_lb[k]=Double();
				func_table_ub[k]=Double();
				if (store.check(var,k,false)!=INCONSISTENT ){
			    	gm.variables[var]->addr_value()=k;
			    	int address=Variable::getAddress(function.variables());
			    	func_table_ub[k]=function.table()[address];
			  	}
			  	if(store.check(var,k,false)==CONSISTENT){
			    	gm.variables[var]->addr_value()=k;
			    	int address=Variable::getAddress(function.variables());
			    	func_table_lb[k]=function.table()[address];
			  	}
			}
			//Advance next_pointer
			store.check(var,val,true);
			gm.variables[var]->value()=val;
			//Normalize function table
			normalize_table(func_table_lb);
			normalize_table(func_table_ub);
			assert(!func_table_lb[val].isZero());
			assert(!func_table_ub[val].isZero());
			sample_weight_lb+=log(func_table_lb[val].value());
			sample_weight_ub+=log(func_table_ub[val].value());
			
		}
		
		//cout<<all_weights[i]<<" "<<sample_weight<<" "<<exp(all_weights[i]-sample_weight)<<endl;
		overall_weight_lb+=exp(all_weights[i]-sample_weight_lb);
		overall_weight_ub+=exp(all_weights[i]-sample_weight_ub);
	}
	for(int j=0;j<sampling_order.size();j++){
		  gm.variables[sampling_order[j]]->value()=INVALID_VALUE;
	}
	store.reset();
	return pair<double,double>(overall_weight_lb,overall_weight_ub);
}
Double RBSampleSearch::computePEApp(GM& gm, JG& jg, int p_bound, vector<int>& order,vector<int>& sampling_order, int num_samples, int max_restarts)
{
	// Initialize the SAT solver
	gm.convertToSAT();
	Solver solver;
	gzFile in = gzopen(satfile, "rb");
	parse_DIMACS(in,solver);
	gzclose(in);
	
	vector<bool> is_sampled(order.size());
	for(int i=0;i<sampling_order.size();i++)
		is_sampled[sampling_order[i]]=true;
	vector<int> helper_order;
	vector<int> helper_sampling_order;
	
	for(int i=0;i<order.size();i++)
		if(!is_sampled[order[i]])
			helper_order.push_back(order[i]);
		else
			helper_sampling_order.push_back(order[i]);
	//for(int i=(int)sampling_order.size()-1;i>-1;i--)
	//	helper_order.push_back(sampling_order[i]);
	for(int i=0;i<helper_sampling_order.size();i++)
		helper_order.push_back(helper_sampling_order[i]);
	sampling_order=vector<int>();
	for(int i=helper_sampling_order.size()-1;i>-1;i--)
		sampling_order.push_back(helper_sampling_order[i]);
	assert(helper_order.size()==order.size());
	IJGPSamplingHelper helper(&gm,&jg,p_bound,helper_order);
	myRandom random;
	double weight_lb=0.0,weight_ub=0.0;
	time_t curr_time;
	
	// Initialize the sampling order
	ORStore store(gm,sampling_order);
	vector<vector<int> > all_samples;
	vector<double> all_weights;
	
	for(int i=0;i<num_samples;i++)
	{
		// Set all variables to invalid value
		for(int j=0;j<sampling_order.size();j++)
			gm.variables[sampling_order[j]]->value()=INVALID_VALUE;

		vec<Lit> assumptions;
		all_samples.push_back(vector<int>());
		int cs=all_samples.size()-1;
		double sample_weight=0.0;
		assert(store.next_pointer == 0);
		// Sample all variables according to the sampling order
		for(int j=0;j<sampling_order.size();j++){
			int var=sampling_order[j];
			vector<Double> func_table(gm.variables[var]->domain_size());
			assert(gm.variables[var]->value()==INVALID_VALUE);
			// Initialize the proposal distribution for sampling
			Function& function=helper.getFunction(var);
			for(int k=0;k<gm.csp_to_sat_variables[var].size();k++)
			{
				gm.variables[var]->addr_value()=k;
				int address=Variable::getAddress(function.variables());
				func_table[k]=function.table()[address];
			}
			
			// Update the OR store
			for(int k=0;k<gm.csp_to_sat_variables[var].size();k++)
			{
				if (store.check(var,k,false)==DONTKNOW){
					assumptions.push(Lit(gm.csp_to_sat_variables[var][k]));
					lbool solver_status = solver.solve(max_restarts,assumptions);
					if(solver_status == l_True)
						store.AddGood(var,k);
					else if (solver_status == l_False){
						store.AddNoGood(var,k);
					}
					assumptions.pop();
				}
			}
			
			while(1){
				int value=generateSample(func_table,random,true);
				assert(value!=INVALID_VALUE);
				assumptions.push(Lit(gm.csp_to_sat_variables[var][value]));
				if (isConsistent(var,value,assumptions,store,solver)){
					all_samples[cs].push_back(value);
					gm.variables[var]->value()=value;
					break;
				}
				else{
					func_table[value]=Double();
					assumptions.pop();
				}
			}
		}
		BE be(gm.variables,gm.functions,order);
		double actual_weight=log(be.log_pe.toDouble().value());
		all_weights.push_back(actual_weight);
		for(int j=0;j<gm.variables.size();j++){
		  gm.variables[j]->value()=INVALID_VALUE;
		}
		cout<<"Sample "<<i<<"generated\n";
		//Output z value after 10000 samples are generated
		if (i > 1 && i % 100 == 0){
			pair<double,double> tmp;
			tmp=purgeORStore(gm, helper, store, sampling_order, all_samples, all_weights);
			weight_lb+=tmp.first;
			weight_ub+=tmp.second;
			//weight+=(exp(actual_weight-sample_weight));
			Double pe_lb=Double(weight_lb/(double)i+1);
			pe_lb*=gm.mult_factor;
			Double pe_ub=Double(weight_ub/(double)i+1);
			pe_ub*=gm.mult_factor;
			//cout.setf(ios::fixed,ios::floatfield);
			time_t some_time=time(NULL);
			//if(curr_time != some_time){
			//curr_time=time(NULL);
			cout<<"z <"<<log10(pe_lb.value())<<","<< log10(pe_ub.value())<<" > "<<endl;
			cout<<"z <"<<pe_lb.value()<<","<< pe_ub.value()<<" > "<<endl;
			all_samples=vector<vector<int> > ();
			all_weights=vector<double>();
		}
	}
		
	for(int j=0;j<gm.variables.size();j++)
	{
		gm.variables[j]->value()=INVALID_VALUE;
	}
	return Double(weight_ub)/Double((double)num_samples);
}
	
void RBSampleSearch::computeBeliefs(GM& gm, JG& jg, int p_bound, vector<set<int> >& clusters,vector<int>& order,vector<int>& sampling_order, int num_samples)
{
	gm.convertToSAT();

	Solver solver;
	gzFile in = gzopen(satfile, "rb");
	parse_DIMACS(in,solver);
	gzclose(in);
	
	cerr<<"Solver inited\n";

	vector<bool> is_sampled(order.size());
	for(int i=0;i<sampling_order.size();i++)
		is_sampled[sampling_order[i]]=true;
	vector<int> helper_order;
	vector<int> helper_sampling_order;
	vector<int> other_variables;
	
	for(int i=0;i<order.size();i++)
		if(!is_sampled[order[i]])
		{
			helper_order.push_back(order[i]);
			other_variables.push_back(order[i]);
		}
		else
			helper_sampling_order.push_back(order[i]);
	//for(int i=(int)sampling_order.size()-1;i>-1;i--)
	//	helper_order.push_back(sampling_order[i]);
	for(int i=0;i<helper_sampling_order.size();i++)
		helper_order.push_back(helper_sampling_order[i]);
	sampling_order=vector<int>();
	for(int i=helper_sampling_order.size()-1;i>-1;i--)
		sampling_order.push_back(helper_sampling_order[i]);
	assert(helper_order.size()==order.size());
	IJGPSamplingHelper helper(&gm,&jg,p_bound,helper_order);
	myRandom random;
	Double weight;
	vector<vector<Double> > marginals(gm.variables.size());
	for(int i=0;i<gm.variables.size();i++)
		marginals[i]=vector<Double>(gm.variables[i]->domain_size());
	JT jt(gm.variables,gm.functions,clusters,order);
	time_t curr_time;
	for(int i=0;i<num_samples;i++)
	{
		for(int j=0;j<sampling_order.size();j++)
			gm.variables[j]->value()=INVALID_VALUE;

		vec<Lit> assumptions;
		double sample_weight=0.0;
		for(int j=0;j<sampling_order.size();j++)
		{
			int var=sampling_order[j];
			vector<Double> func_table(gm.variables[var]->domain_size());
			Function& function=helper.getFunction(var);
			//if(j==0)
			//	cerr<<"tab-size = "<<function.table().size()<<endl;
			for(int k=0;k<gm.csp_to_sat_variables[var].size();k++)
			{
				gm.variables[var]->addr_value()=k;
				int address=Variable::getAddress(function.variables());
				func_table[k]=function.table()[address];
			}
			for(int k=0;k<gm.csp_to_sat_variables[var].size();k++)
			{
				assumptions.push((Lit(gm.csp_to_sat_variables[var][k])));
				if(!solver.solve(assumptions))
				{
					func_table[k]=Double();
				}
				//cout<<"Done\n";
				assumptions.pop();
			}
			Double norm_const;
			// Normalize the fun_table
			for(int k=0;k<func_table.size();k++)
			{
				norm_const+=func_table[k];
			}
			for(int k=0;k<func_table.size();k++)
			  func_table[k]/=norm_const;
			int value=INVALID_VALUE;
			double rand_num=random.getDouble();
			double cdf=0.0;
			double curr_var_weight=0.0;
			for(int k=0;k<func_table.size();k++)
			{
				if(!func_table[k].isZero())
				{
					value=k;
					cdf+=func_table[k].value();
					curr_var_weight=func_table[k].value();
					if(rand_num <= cdf)
					{
						value=k;
						break;
					}
				}
			}
			assert(value!=INVALID_VALUE);
			gm.variables[var]->value()=value;
			assumptions.push((Lit(gm.csp_to_sat_variables[var][value])));
			sample_weight+=log(curr_var_weight);
		}
		
		jt.propagate();
		BE be(gm.variables,gm.functions,order);
		double actual_weight=log(be.log_pe.toDouble().value());
		weight=Double(exp(actual_weight-sample_weight));
		
		vector<Function>& marginals_=jt.marginals;
		// Update marginals
		for(int j=0;j<sampling_order.size();j++)
		{
			int id=sampling_order[j];
			marginals[id][gm.variables[id]->value()]+=weight;
		}
		for(int j=0;j<other_variables.size();j++)
		{
			int id=other_variables[j];
			for(int k=0;k<gm.variables[id]->domain_size();k++)
			{
				marginals[id][k]+=marginals_[id].table()[k];
			}
		}
		for(int j=0;j<gm.variables.size();j++)
		{
			gm.variables[j]->value()=INVALID_VALUE;
		}
		time_t some_time=time(NULL);
		if(curr_time != some_time){
		  curr_time=time(NULL);
		  gm.printMarginalsUAI08(marginals);
		}
	}
}

bool SampleSearch::isconsistent(GM& gm)
{
  gm.convertToSAT();
  
  Solver solver;
  gzFile in = gzopen(satfile, "rb");
  parse_DIMACS(in,solver);
  gzclose(in);
  return solver.solve();
}

//void RBSampleSearch::addTrace(GM &gm, std::vector<int> &sampling_order, ORStore &store, vec<Lit> &trail, vec<int> trail_lim)
//{
//
//}
pair<double,double> purgeORStoreSAT(GM& gm, IJGPSamplingHelper& helper, ORStore& store, vector<int>& sampling_order, vector<vector<int> >& all_samples, vector<double>& all_weights)
{
    
	assert(all_samples.size()==all_weights.size());
	double overall_weight_ub=0.0;
	double overall_weight_lb=0.0;
	for(int i=0;i<all_samples.size();i++){
		double sample_weight_lb=0.0,sample_weight_ub=0.0;
		for(int j=0;j<sampling_order.size();j++){
		  gm.variables[sampling_order[j]]->value()=INVALID_VALUE;
		}
		assert(sampling_order.size() == all_samples[i].size());
		for(int j=0;j<all_samples[i].size();j++){
			int var=sampling_order[j];
			int val=all_samples[i][j];
			vector<Double> func_table_lb(gm.variables[var]->domain_size());
			vector<Double> func_table_ub(gm.variables[var]->domain_size());
			Function& function=helper.getFunction(var);
			for(int k=0;k<2;k++)
			{
				func_table_lb[k]=Double();
				func_table_ub[k]=Double();
				if (store.check(var,k,false)!=INCONSISTENT ){
			    	gm.variables[var]->addr_value()=k;
			    	int address=Variable::getAddress(function.variables());
			    	func_table_ub[k]=function.table()[address];
			  	}
			  	if(store.check(var,k,false)==CONSISTENT){
			    	gm.variables[var]->addr_value()=k;
			    	int address=Variable::getAddress(function.variables());
			    	func_table_lb[k]=function.table()[address];
			  	}
			}
			//Advance next_pointer
			store.check(var,val,true);
			gm.variables[var]->value()=val;
			//Normalize function table
			normalize_table(func_table_lb);
			normalize_table(func_table_ub);
			assert(!func_table_lb[val].isZero());
			assert(!func_table_ub[val].isZero());
			sample_weight_lb+=log(func_table_lb[val].value());
			sample_weight_ub+=log(func_table_ub[val].value());
			
		}
		
		//cout<<all_weights[i]<<" "<<sample_weight<<" "<<exp(all_weights[i]-sample_weight)<<endl;
		overall_weight_lb+=exp(all_weights[i]-sample_weight_lb);
		overall_weight_ub+=exp(all_weights[i]-sample_weight_ub);
	}
	for(int j=0;j<sampling_order.size();j++){
		  gm.variables[sampling_order[j]]->value()=INVALID_VALUE;
	}
	//store.reset();
	return pair<double,double>(overall_weight_lb,overall_weight_ub);
}
RBSampleSearchSAT::RBSampleSearchSAT(): interval(10)
{
}

Double RBSampleSearchSAT::computePEApp(GM&gm, JG& jg, const char* satfile, int p_bound, vector<int>& order,vector<int>& sampling_order, int num_samples, int max_restarts)
{
	// Initialize the SAT solver
	Solver solver;
	gzFile in = gzopen(satfile, "rb");
	parse_DIMACS(in,solver);
	gzclose(in);
	
	vector<bool> is_sampled(order.size());
	for(int i=0;i<sampling_order.size();i++)
		is_sampled[sampling_order[i]]=true;
	vector<int> helper_order;
	vector<int> helper_sampling_order;
	
	for(int i=0;i<order.size();i++)
		if(!is_sampled[order[i]])
			helper_order.push_back(order[i]);
		else
			helper_sampling_order.push_back(order[i]);
	//for(int i=(int)sampling_order.size()-1;i>-1;i--)
	//	helper_order.push_back(sampling_order[i]);
	for(int i=0;i<helper_sampling_order.size();i++)
		helper_order.push_back(helper_sampling_order[i]);
	sampling_order=vector<int>();
	for(int i=helper_sampling_order.size()-1;i>-1;i--)
		sampling_order.push_back(helper_sampling_order[i]);
	assert(helper_order.size()==order.size());
	IJGPSamplingHelper helper(&gm,&jg,p_bound,helper_order);
	myRandom random;
	double weight_lb=0.0,weight_ub=0.0;
	time_t curr_time;
	
	// Initialize the sampling order
	ORStore store(gm,sampling_order);
	vector<vector<int> > all_samples;
	vector<double> all_weights;
	//double tmp_exact_weight=0.0;
	
	for(int i=0;i<num_samples;i++)
	{
		// Set all variables to invalid value
		for(int j=0;j<gm.variables.size();j++)
			gm.variables[j]->value()=INVALID_VALUE;

		vec<Lit> assumptions;
		all_samples.push_back(vector<int>());
		int cs=all_samples.size()-1;
		double sample_weight=0.0;
		double tmp_sample_weight=1.0;
		assert(store.next_pointer == 0);
		vec<Lit> tmp_assumptions;
		set<Lit> inferred_lits;
		// Sample all variables according to the sampling order
		for(int j=0;j<sampling_order.size();j++){
			int var=sampling_order[j];
			vector<Double> func_table(gm.variables[var]->domain_size());
			//assert(gm.variables[var]->value()==INVALID_VALUE);
			//if (gm.variables[var]->value()!=INVALID_VALUE){
			//	Lit q= (gm.variables[var]->value()==0)?(~Lit(var)):(Lit(var));
			//	int other_val=(gm.variables[var]->value()==0)?(1):(0);
//				if (inferred_lits.find(q)!=inferred_lits.end()){
//					store.AddNoGood(var,other_val);
//					store.AddAssignment(var,gm.variables[var]->value());
//				}
//				continue;
//			}
			// Initialize the proposal distribution for sampling
			Function& function=helper.getFunction(var);
			//vector<double> tmp_table(2);
			for(int k=0;k<2;k++)
			{
				gm.variables[var]->addr_value()=k;
				int address=Variable::getAddress(function.variables());
				func_table[k]=function.table()[address];
			}
			//for(int k=0;k<2;k++){
			//	tmp_table[k]=(func_table[k].value())/(func_table[0].value()+func_table[1].value());
			//}
			// Update the OR store
			if (store.check(var,0,false)==DONTKNOW){
				assumptions.push(~Lit(var));
				lbool solver_status = solver.solve(max_restarts,assumptions);
				if(solver_status == l_True)
					store.AddGood(var,0);
				else if (solver_status == l_False){
					store.AddNoGood(var,0);
					store.AddGood(var,1);
				}
				assumptions.pop();
			}
			if (store.check(var,1,false)==DONTKNOW){
				assumptions.push(Lit(var));
				lbool solver_status = solver.solve(max_restarts,assumptions);
				if(solver_status == l_True)
					store.AddGood(var,1);
				else if (solver_status == l_False){
					store.AddNoGood(var,1);
					store.AddGood(var,0);
				}
				assumptions.pop();
			}
			int value=-1;
			while(1){
				value=generateSample(func_table,random,true);
				assert(value!=INVALID_VALUE);
				if (value==0){
					assumptions.push(~Lit(var));
				}
				else {
					assumptions.push(Lit(var));
				}
				if (isConsistentSAT(gm,var,value,assumptions,inferred_lits,store,solver)){
					all_samples[cs].push_back(value);
					gm.variables[var]->value()=value;
					break;
				}
				else{
					func_table[value]=Double();
					assumptions.pop();
				}
			}
			
		//BE be(gm.variables,gm.functions,order);
			BESAT be(gm.variables,gm.clauses,order);
			double actual_weight=be.log_pe;//log(be.log_pe.toDouble().value());
		//tmp_exact_weight+=(exp(be.log_pe)/tmp_sample_weight);
		//cout<<actual_weight<<" [ "<<flush;
		//cout<<log(checkNumSolutions(gm,assumptions))<<" ]"<<flush;
		//double actual_weight=log(checkNumSolutions(gm,assumptions));
		//cout<<"("<<actual_weight<<" , "<<be.log_pe<<" ) "<<flush;
		//assert(fabs(actual_weight-be.log_pe) < 0.001);
			all_weights.push_back(actual_weight);
			for(int j=0;j<gm.variables.size();j++){
		  	gm.variables[j]->value()=INVALID_VALUE;
			}
			cout<<"Sample "<<i<<" generated\n";
		//Output z value after interval samples are generated
			if (i > 1 && i % interval == 0){
				pair<double,double> tmp;
				tmp=purgeORStoreSAT(gm, helper, store, sampling_order, all_samples, all_weights);
			//weight_lb+=tmp.first;
			//weight_ub+=tmp.second;
			//weight+=(exp(actual_weight-sample_weight));
				Double pe_lb=Double((weight_lb+tmp.first)/(double)i+1);
				pe_lb*=gm.mult_factor;
				Double pe_ub=Double((weight_ub+tmp.second)/(double)i+1);
				pe_ub*=gm.mult_factor;
			//cout.setf(ios::fixed,ios::floatfield);
			//if(curr_time != some_time){
			//curr_time=time(NULL);
			//cout<<"z <"<<log10(pe_lb.value())<<","<< log10(pe_ub.value())<<" > "<<endl;
				cout<<"z <"<<pe_lb.value()<<","<< pe_ub.value()<<" > "<<" , ";
				cout<<"Storage space = "<<printbytes(store.storage.size()*sizeof(ORStoreNode))<<endl;
			//cout<< "Exact z = "<<tmp_exact_weight/(double)i+1<<endl;
				if (tmp.first==tmp.second){
					all_samples=vector<vector<int> > ();
					all_weights=vector<double>();
					weight_lb+=tmp.first;
					weight_ub+=tmp.second;
					tmp.first=0.0;
					tmp.second=0.0;
				}
				if ((int)store.storage.size() > 234217728){
					all_samples=vector<vector<int> > ();
					all_weights=vector<double>();
					weight_lb+=tmp.first;
					weight_ub+=tmp.second;
					store.reset();
				} 
			}
		}
		
		for(int j=0;j<gm.variables.size();j++)
		{
			gm.variables[j]->value()=INVALID_VALUE;
		}
		return Double(weight_ub)/Double((double)num_samples);
	}
}

double RBSampleSearchSAT::runBE(GM& gm, vector<int>& order, char* infile)
{
	Solver solver;
	gzFile in = gzopen(infile, "rb");
	parse_DIMACS(in,solver);
	gzclose(in);
	
	
	//assumptions.push(Lit(0));
	BESAT be_sat(gm.variables,gm.clauses,order);
	cout<<be_sat.log_pe<<endl;
	return exp(be_sat.log_pe); 
}