#include "GM.h"
#include "JG.h"
#include "DRC.h"
//bool less_than_comparator_variable(const Variable* a, const Variable* b)
//{
//	if(a->id()<b->id())
//		return true;
//	return false;
//}
void Function::readCNF(istream& in, vector<Variable*>& all_variables)
{
	do{
		int curr_var;
		in>>curr_var;
		if(curr_var ==0)
			break;
		int var_id=((curr_var > 0)? (curr_var-1):(-curr_var-1));
		variables_.push_back(all_variables[var_id]);
		if(curr_var>0)
		{
			all_variables[var_id]->addr_value()=0;
		}
		else
		{
			all_variables[var_id]->addr_value()=1;
		}
		
	}while(1);
	
	sort(variables_.begin(),variables_.end(),less_than_comparator_variable);
	
	int table_size=Variable::getDomainSize(variables_);
	table_=vector<Double> (table_size);
	for(int i=0;i<table_.size();i++)
	{
		table_[i]=Double(1.0);
	}
	
	table_[Variable::getAddress(variables_)]=Double();
}

bool GM::readCNF(char* infilename)
{
	ifstream xInputFile(infilename);
	char cCheck;
	char buffer[5000];
	int num_variables,num_constraints;
	while (1) 
	{
		xInputFile >> cCheck;
		if (cCheck == 'c') 
		{
			xInputFile.getline(buffer, 5000);
			continue;
		}
		else if (cCheck == 'p') 
		{
			xInputFile >> buffer;
			xInputFile >> num_variables;
			xInputFile >> num_constraints;
			break;
		}
		else 
		{
			cout << "\nError: File not in DIMACS format?\n";
			//return 0;
			return false;
		}
	}
	// Read the variables
	variables=vector<Variable*>(num_variables);
	for(int i=0;i<num_variables;i++)
	{
		variables[i]=new Variable(i,2);
	}
	functions=vector<Function*>(num_constraints);
	for(int i=0;i<num_constraints;i++)
		functions[i]=new Function(i);
	int i=0;
	while (1) 
	{
		if(i==num_constraints)
			break;
		if (xInputFile.eof()) 
		{
			cout << "\nError: Unexpected end of file.\n";
			//return 0;
			exit(1);
		}
		xInputFile >> cCheck;
		if (cCheck == 'c') 
		{
			xInputFile.getline(buffer, 5000);
			continue;
		}
		else xInputFile.putback(cCheck);
		functions[i]->readCNF(xInputFile,variables);
		i++;
	}
	return true;
}
void GM::convertToSAT()
{

	// First map the variables between SAT and the belief network
	// For each variable we create a clause
	// Create a value for each variable value assignment
	int count=0;
	csp_to_sat_variables=vector<vector<int> > (variables.size());
	int num_con=0;
	for(int i=0;i<variables.size();i++)
	{
		for(int j=0;j<variables[i]->domain_size();j++)
		{
			sat_to_csp_variables.push_back(i);
			csp_to_sat_variables[i].push_back(count);
			count++;
		}
		/*num_con+=combination(variables[i]->num_values,2);
		if(variables[i]->num_values >1)
			num_con++;*/
	}
	for(int i=0;i<variables.size();i++)
	{
		num_con++;
		for(int j=0;j<csp_to_sat_variables[i].size();j++)
		{
			for(int k=j+1;k<csp_to_sat_variables[i].size();k++)
			{
				num_con++;
			}
		}
	}
	//// Add all evidence variables as clauses
	//for(int i=0;i<num_variables;i++)
	//	if(variables[i]->is_evidence)
	//		num_con++;
	/*for( int i=0;i<num_constraints;i++)
	{
		for(int j=0;j<constraints[i]->table.size();j++)
		{
			Variable::setAddressValues(constraints[i]->scope,j);
			if(!constraints[i]->table[constraints[i]->getEntry()])
			{
				num_con++;
			}
		}
	}*/
	//num_con+=num_constraints;
	cerr<<"Counting constraints\n";
	//Find the number of constraints that can be derived from the cpts
	for(int i=0;i<functions.size();i++)
	{
		Function* function =functions[i];
		for(int j=0;j<function->table().size();j++)
		{
			if(function->table()[j].isZero())
				num_con++;
		}
	}
	//cerr<<"Writing to file\n";
	ofstream outfile ("temp.cnf");
	outfile<<"p cnf "<<sat_to_csp_variables.size()<<" "<<num_con<<endl;

	//Now write constraints to ostream
	for(int i=0;i<variables.size();i++)
	{
		
		
		// A constraint making sure that each variable is assigned a value
		for(int j=0;j<csp_to_sat_variables[i].size();j++)
		{
			outfile<<csp_to_sat_variables[i][j]+1<<" ";
		}
		outfile<<"0 \n";
		// A constraint making sure that each variable is assigned a value
		for(int j=0;j<csp_to_sat_variables[i].size();j++)
		{
			for(int k=j+1;k<csp_to_sat_variables[i].size();k++)
			{
				outfile<<-(csp_to_sat_variables[i][j]+1)<<" "<<-(csp_to_sat_variables[i][k]+1)<<" 0 "<<endl;
			}
		}

	}
	for(int i=0;i<functions.size();i++)
	{
		Function* function = functions[i];
		
		for(int j=0;j<function->table().size();j++)
		{
			Variable::setAddress(function->variables(),j);
			if(function->table()[j].isZero())
			{
			
				for(int k=0;k<function->variables().size();k++)
				{
					/*if(function->scope[k]->addr_value==0)
					{
						outfile<<(function->scope[k]->id+1)<<" ";
					}
					else
					{
						outfile<<-(function->scope[k]->id+1)<<" ";
					}*/
					outfile<<-(csp_to_sat_variables[function->variables()[k]->id()][function->variables()[k]->addr_value()]+1)<<" ";
				}
				outfile<<"0\n";
			}
		}
	}

	// Write evidence
	/*for(int i=0;i<num_variables;i++)
		if(variables[i]->is_evidence)
		{
			assert(variables[i]->value!=INVALID_VALUE);
			outfile<<csp_to_sat_variables[i][variables[i]->value]+1<<" 0\n";
		}*/
	/*for( int i=0;i<num_constraints;i++)
	{
		for(int j=0;j<constraints[i]->table.size();j++)
		{
			Variable::setAddressValues(constraints[i]->scope,j);
			if(!constraints[i]->table[constraints[i]->getEntry()])
			{
				for(int k=0;k<constraints[i]->scope_size;k++)
				{
					int var=constraints[i]->scope[k]->id;
					int val=constraints[i]->scope[k]->addr_value;
					outfile<<-(csp_to_sat_variables[var][val]+1)<<" ";
				}
				outfile<<" 0\n";
			}
		}
	}*/
	outfile.close();

}

bool GM::readVIB(char* bayes_file)
{
	cout<<"Reading the Bayesian Network..."<<flush;
	//Open the file
	ifstream infile(bayes_file);
	if(!infile.good())
	{
		cerr<<"Empty Bayesnet file\n";
		return false;
	}
	int num_variables;
	// Read number of variables
	infile>>num_variables;
	//Read the values of each variable
	variables=vector<Variable*>(num_variables);
	
	//Read variable information
	for(int i=0;i<num_variables;i++)
	{
		int num_values;
		infile>>num_values;
		vector<int> domain;
		for(int j=0;j<num_values;j++)
		{
			domain.push_back(j);
		}
		variables[i]=new Variable(i,domain);
	}
	
	int num_cpts;
	//Read #conditional probability tables
	infile>>num_cpts;
	
	functions=vector<Function*> (num_cpts);

	// Currently we only allow num cpts = numvariables
	assert(num_cpts==num_variables);
	int max_parents=0;
	//Read the cpts
	for(int i=0;i<num_cpts;i++)
	{
		functions[i]=new CPT();
		CPT* cpt=dynamic_cast<CPT*> (functions[i]);
		cpt->id()=i;
		
		//Read the variables in the function
		int num_variables_infunc;
		infile>>num_variables_infunc;
		int marg_variable;
		infile>>marg_variable;
		cpt->setMargVariable(variables[marg_variable]);
		//func_variables.push_back(variables[marg_variable]);
		for(int j=1;j<num_variables_infunc;j++)
		{
			int temp;
			infile>>temp;
			cpt->cond_variables().push_back(variables[temp]);
		}
		if(max_parents < num_variables_infunc)
			max_parents=num_variables_infunc;
		// Currently in vib format we allow specification of functions in order
		assert(marg_variable==i);

		cpt->variables()=cpt->cond_variables();
		cpt->variables().push_back(cpt->marg_variable());
		sort(cpt->variables().begin(),cpt->variables().end(),less_than_comparator_variable);
		
		//int num_values=Variable::getDomainSize(sorted_func_variables);
		////int marg_num_values=Variable::getNumValues(marg_variables);
		//functions[i]->variables()=sorted_func_variables;
		int cond_num_values=Variable::getDomainSize(cpt->cond_variables());
		int num_values=Variable::getDomainSize(cpt->variables());
		cpt->table()=vector<Double> (num_values);
		
		for(int j=0;j<cond_num_values;j++)
		{
			Variable::setAddressVIB(cpt->cond_variables(),j);
			for(int k=0;k<cpt->marg_variable()->domain_size();k++)
			{
				cpt->marg_variable()->addr_value()=k;
				double value;
				infile>>value;
				
				int entry=Variable::getAddress(cpt->variables());
				if(value > DBL_MIN)
				{		
					cpt->table()[entry]=Double(value);
				}
			}
		}
		sort(cpt->cond_variables().begin(),cpt->cond_variables().end(),less_than_comparator_variable);
		/*if(i<5)
		{
			
			for(int j=0;j<functions[i]->table().size();j++)
			{
				Variable::setAddress(functions[i]->variables(),j);
				cout<<"tab[ ";
				for(int k=0;k<functions[i]->variables().size();k++)
				{
					cout<<functions[i]->variables()[k]->addr_value()<<" ";
				}
				cout<<"] = "<<functions[i]->table()[j]<<endl;
			}
		}*/
	}

	// Construct the graph of the bayesian network
	//cout<<"Max parents = "<<max_parents+1<<endl;
	cout<<"Done\n";
	return true;
}
void GM::readEvidence1(char* evidencefile)
{
	mult_factor=Double((double)1.0);
	int num_evidence;
	ifstream evidfile(evidencefile);
	char buffer[1000];
	evidfile.getline(buffer,1000);
	evidfile>>num_evidence;
	for(int i=0;i<num_evidence;i++)
	{
		int var,val;
		evidfile>>var>>val;
		assert(var < variables.size());
		assert(val < variables[var]->domain_size());
		Function *func=new Function();
		func->variables().push_back(variables[var]);
		func->table()=vector<Double>(variables[var]->domain_size());
		func->table()[val]=Double((double)1.0);
		functions.push_back(func);
	}
}

void GM::readEvidenceErgo(char* evidencefile)
{
	mult_factor=Double((double)1.0);
	vector<Variable*> evidence_variables;

	int num_evidence;
	ifstream evidfile(evidencefile);
	// Just ignore the first line

	char buffer[1000];
	evidfile.getline(buffer,1000);
	evidfile>>num_evidence;
	for(int i=0;i<num_evidence;i++)
	{
		int var,val;
		evidfile>>var>>val;
		evidence_variables.push_back(variables[var]);
		variables[var]->value()=val;
		variables[var]->addr_value()=val;
	}
	sort(evidence_variables.begin(),evidence_variables.end(),less_than_comparator_variable);
	//cerr<<"num-evidence = "<<evidence_variables.size()<<endl;
	vector<int> variable_mapping(variables.size());
	vector<bool> is_evidence(variables.size());
	for(int i=0;i<variables.size();i++)
	{
		is_evidence[i]=false;
	}
	for(int i=0;i<evidence_variables.size();i++)
		is_evidence[evidence_variables[i]->id()]=true;

	int var_count=0;
	for(int i=0;i<variable_mapping.size();i++)
	{
		if(!is_evidence[i])
		{		
			variable_mapping[i]=var_count;
			var_count++;
		}
		else
		{
			variable_mapping[i]=INVALID_VALUE;
		}
	}

	vector<bool> deleted_functions(functions.size());
	//update functions
	for(int i=0;i<functions.size();i++)
	{
		vector<Variable*> func_evid_vars;
		deleted_functions[i]=false;
		do_set_intersection(functions[i]->variables(),evidence_variables,func_evid_vars,less_than_comparator_variable);
		if(func_evid_vars.empty())
			continue;
		// All variables are evidence variables
		if(func_evid_vars.size()==functions[i]->variables().size())
		{
			int address=Variable::getAddress(functions[i]->variables());
			mult_factor*=functions[i]->table()[address];
			
			deleted_functions[i]=true;
			continue;
		}
		vector<Variable*> other_variables;
		do_set_difference(functions[i]->variables(),func_evid_vars,other_variables,less_than_comparator_variable);
		int other_num_values=Variable::getDomainSize(other_variables);
		vector<Double> new_table(other_num_values);
		for(int j=0;j<other_num_values;j++)
		{
			Variable::setAddress(other_variables,j);
			int entry=Variable::getAddress(functions[i]->variables());
			new_table[j]=functions[i]->table()[entry];
		}
		/*for(int j=0;j<other_variables.size();j++)
		{
			cout<<"Map ="<<variable_mapping[other_variables[j]->id()]<<" \t";
			other_variables[j]->print();
		}
		cout<<"--------\n";*/
		functions[i]->table()=new_table;
		functions[i]->variables()=other_variables;
	}
	//for(int i=0;i<functions.size();i++)
	//{
	//	vector<Variable*> func_evid_vars;
	//	//deleted_functions[i]=false;
	//	do_set_intersection(functions[i]->variables(),evidence_variables,func_evid_vars,less_than_comparator_variable);
	//	if(!func_evid_vars.empty())
	//	{
	//		cerr<<"Something wrong in assigning evidence\n";
	//	}
	//}
	cerr<<"Mult-factor = "<<mult_factor<<endl;
	//update functions;
	vector<Function*> new_functions;
	int count=0;
	for(int i=0;i<functions.size();i++)
	{
		if(deleted_functions[i])
		{
			Function* func_pt=functions[i];
			functions[i]=NULL;
			delete(func_pt);
		}
		else
		{
			functions[i]->id()=count++;
			new_functions.push_back(functions[i]);
		}
	}
	functions=new_functions;
	//update variables
	vector<Variable*> new_variables;
	var_to_ids=vector<int> (variables.size()-num_evidence);
	for(int i=0;i<variables.size();i++)
	{
		if(is_evidence[i])
		{
			Variable* var_pt=variables[i];
			variables[i]==NULL;
			delete(var_pt);
		}
		else
		{
			variables[i]->id()=variable_mapping[i];
			var_to_ids[variables[i]->id()]=i;
			new_variables.push_back(variables[i]);
		}
	}
	variables=new_variables;
	//for(int i=0;i<variables.size();i++)
	//	cout<<variables[i]->id()<<" ";
	//cout<<endl;
}

void GM::readEvidenceVIB(char* evidencefile)
{
	mult_factor=Double((double)1.0);
	vector<Variable*> evidence_variables;

	int num_evidence;
	ifstream evidfile(evidencefile);
	// Just ignore the first line

	char buffer[1000];
	//evidfile.getline(buffer,1000);
	evidfile>>num_evidence;
	for(int i=0;i<num_evidence;i++)
	{
		int var,val;
		evidfile>>var>>val;
		evidence_variables.push_back(variables[var]);
		cout<<var<<" "<<val<<endl;
		variables[var]->value()=val;
		variables[var]->addr_value()=val;
	}
	sort(evidence_variables.begin(),evidence_variables.end(),less_than_comparator_variable);
	//cerr<<"num-evidence = "<<evidence_variables.size()<<endl;
	vector<int> variable_mapping(variables.size());
	vector<bool> is_evidence(variables.size());
	for(int i=0;i<variables.size();i++)
	{
		is_evidence[i]=false;
	}
	for(int i=0;i<evidence_variables.size();i++)
		is_evidence[evidence_variables[i]->id()]=true;

	int var_count=0;
	for(int i=0;i<variable_mapping.size();i++)
	{
		if(!is_evidence[i])
		{		
			variable_mapping[i]=var_count;
			var_count++;
		}
		else
		{
			variable_mapping[i]=INVALID_VALUE;
		}
	}

	vector<bool> deleted_functions(functions.size());
	//update functions
	for(int i=0;i<functions.size();i++)
	{
		vector<Variable*> func_evid_vars;
		deleted_functions[i]=false;
		do_set_intersection(functions[i]->variables(),evidence_variables,func_evid_vars,less_than_comparator_variable);
		if(func_evid_vars.empty())
			continue;
		// All variables are evidence variables
		if(func_evid_vars.size()==functions[i]->variables().size())
		{
			int address=Variable::getAddress(functions[i]->variables());
			mult_factor*=functions[i]->table()[address];
			
			deleted_functions[i]=true;
			continue;
		}
		vector<Variable*> other_variables;
		do_set_difference(functions[i]->variables(),func_evid_vars,other_variables,less_than_comparator_variable);
		int other_num_values=Variable::getDomainSize(other_variables);
		vector<Double> new_table(other_num_values);
		for(int j=0;j<other_num_values;j++)
		{
			Variable::setAddress(other_variables,j);
			int entry=Variable::getAddress(functions[i]->variables());
			new_table[j]=functions[i]->table()[entry];
		}
		/*for(int j=0;j<other_variables.size();j++)
		{
			cout<<"Map ="<<variable_mapping[other_variables[j]->id()]<<" \t";
			other_variables[j]->print();
		}
		cout<<"--------\n";*/
		functions[i]->table()=new_table;
		functions[i]->variables()=other_variables;
	}
	for(int i=0;i<functions.size();i++)
	{
		vector<Variable*> func_evid_vars;
		//deleted_functions[i]=false;
		do_set_intersection(functions[i]->variables(),evidence_variables,func_evid_vars,less_than_comparator_variable);
		if(!func_evid_vars.empty())
		{
			cerr<<"Something wrong in assigning evidence\n";
		}
	}
	cerr<<"Mult-factor = "<<mult_factor<<endl;
	//update functions;
	vector<Function*> new_functions;
	int count=0;
	for(int i=0;i<functions.size();i++)
	{
		if(deleted_functions[i])
		{
			Function* func_pt=functions[i];
			functions[i]=NULL;
			delete(func_pt);
		}
		else
		{
			functions[i]->id()=count++;
			new_functions.push_back(functions[i]);
		}
	}
	functions=new_functions;
	//update variables
	vector<Variable*> new_variables;
	for(int i=0;i<variables.size();i++)
	{
		if(is_evidence[i])
		{
			Variable* var_pt=variables[i];
			variables[i]==NULL;
			delete(var_pt);
		}
		else
		{
			variables[i]->id()=variable_mapping[i];
			new_variables.push_back(variables[i]);
		}
	}
	variables=new_variables;
	//for(int i=0;i<variables.size();i++)
	//	cout<<variables[i]->id()<<" ";
	//cout<<endl;
}

void GM::writeConstraintGraph(char* cgraph)

{
	ofstream outfile(cgraph);

	vector<vector<bool> > adjgraph(this->variables.size());
	for(int i=0;i<variables.size();i++)
	{
		adjgraph[i]=vector<bool> (variables.size());
	}

	int num_edges=0;
	//Now go through each function
	for(int i=0;i<functions.size();i++)
	{
		for(int j=0;j<functions[i]->table().size();j++)
		{
			if(functions[i]->table()[j].isZero())
			{
				for(int k=0;k<functions[i]->variables().size();k++)
				{
					int var1=functions[i]->variables()[k]->id();
					for(int o=k+1;o<functions[i]->variables().size();o++)
					{
						int var2=functions[i]->variables()[o]->id();
						adjgraph[var1][var2]=true;
						adjgraph[var2][var1]=true;
						num_edges++;
					}
				}
				break;
			}
		}
	}

	//Write graph to the file
	outfile<<variables.size()<<" "<<num_edges<<endl;
	for(int i=0;i<variables.size();i++)
	{
		for(int j=i+1;j<variables.size();j++)
		{
			if(adjgraph[i][j])
				outfile<<i+1<<" "<<j+1<<endl;
		}
	}
}

void GM::readErgo(char* infilename)
{
	cerr<<"Reading Bayesnet.."<<flush;
	ifstream infile(infilename);
	char check_char,buffer[1000];
	int num_variables;
	// Read num variables
	while(!infile.eof())
	{
		infile>>check_char;
		if(check_char == '/')
		{
			infile.getline(buffer,1000);
			continue;
		}
		infile.putback(check_char);
		infile>>num_variables;
		break;
	}
	// Read domains
	variables=vector<Variable*> (num_variables);
	int var=0;
	while(!infile.eof())
	{
		infile>>check_char;
		if(check_char == '/')
		{
			infile.getline(buffer,1000);
			continue;
		}
		infile.putback(check_char);
		int domain_size;
		infile>>domain_size;
		vector<int> domain(domain_size);
		for(int i=0;i<domain_size;i++)
			domain[i]=i;
		variables[var]=new Variable(var,domain);
		++var;
		if(var==num_variables)
			break;
	}

	// Read parents of variables
	vector<vector<Variable*> > parents (num_variables);
	var=0;
	while(!infile.eof())
	{
		infile>>check_char;
		if(check_char == '/')
		{
			infile.getline(buffer,1000);
			continue;
		}
		infile.putback(check_char);
		int num_parents;
		infile>>num_parents;
		if(num_parents>0)
		{
			parents[var]=vector<Variable*>(num_parents);
		}
		for(int i=0;i<num_parents;i++)
		{
			int var_num;
			infile>>var_num;
			parents[var][i]=variables[var_num];
		}
		++var;
		if(var==num_variables)
			break;
	}

	// Read the probabilities
	var=0;
	functions=vector<Function*> (num_variables);
	cout<<"\tReading cpts"<<flush;
	while(!infile.eof())
	{
		infile>>check_char;
		if(check_char == '/')
		{
			infile.getline(buffer,1000);
			continue;
		}
		infile.putback(check_char);
		int num_probabilities;
		infile>>num_probabilities;
		
		functions[var]=new CPT();
		CPT* cpt=dynamic_cast<CPT*> (functions[var]);
		cpt->id()=var;
		
		
		cpt->setMargVariable(variables[var]);
		//func_variables.push_back(variables[marg_variable]);
		
		cpt->cond_variables()=parents[var];
		cpt->variables()=cpt->cond_variables();
		cpt->variables().push_back(cpt->marg_variable());
		sort(cpt->variables().begin(),cpt->variables().end(),less_than_comparator_variable);
		
		//int num_values=Variable::getDomainSize(sorted_func_variables);
		////int marg_num_values=Variable::getNumValues(marg_variables);
		//functions[i]->variables()=sorted_func_variables;
		int cond_num_values=Variable::getDomainSize(parents[var]);
		assert(num_probabilities == (cond_num_values*cpt->marg_variable()->domain_size()));
		cpt->table()=vector<Double> (num_probabilities);
		for(int j=0;j<cond_num_values;j++)
		{
			Variable::setAddressVIB(cpt->cond_variables(),j);
			for(int k=0;k<cpt->marg_variable()->domain_size();k++)
			{
				cpt->marg_variable()->addr_value()=k;
				double value;
				infile>>value;
				
				int entry=Variable::getAddress(cpt->variables());
				if(value > DBL_MIN)
				{		
					cpt->table()[entry]=Double(value);
				}
			}
		}
		
		sort(cpt->cond_variables().begin(),cpt->cond_variables().end(),less_than_comparator_variable);
		++var;
		if(var==num_variables)
			break;
	}
	infile.close();
	cerr<<"Done\n";
}
void GM::convertToCN(GM& gm)
{
	gm.variables=this->variables;
	for(int i=0;i<functions.size();i++)
	{
		Function* f = functions[i];
		for(int j=0;j<f->table().size();j++)
		{
			if(f->table()[j].isZero())
			{
				gm.functions.push_back(f);
				break;
			}
		}
	}
}

// Reduce the Bayesian network by constraint processing
void GM::reduceBN(GM* gm, int i_bound, vector<int>& order)
{
	GM csp;
	convertToCN(csp);
	DRC cp_algo(csp,i_bound,order);
	
	// Do domain consistency
	/*
	for(int i=0;i<variables.size();i++)
	{
		vector<bool> domains;
		cp_algo.domainConsistency(i,domains);
	}
	*/
	for(int i=0;i<functions.size();i++)
	{
		cp_algo.reduceFunction(functions[i]);
	}
}

void GM::writeErgo(char* outfile)
{
	ofstream out(outfile);
	assert(functions.size()==variables.size());
	out<<variables.size()<<endl;
	//Write domains
	for(int i=0;i<variables.size();i++)
		out<<variables[i]->domain_size()<<" ";
	out<<endl;
	//Print parents
	for(int i=0;i<functions.size();i++)
	{
		out<<functions[i]->variables().size()-1<<" ";
		bool found=false;
		for(int j=0;j<functions[i]->variables().size();j++)
		{
			if(functions[i]->variables()[j]->id()==i)
			{
				found=true;
			}
			else
			{
				out<<functions[i]->variables()[j]->id()<<" ";
			}
		}
		assert(found==true);
		out<<endl;
	}
	//Write functions
	out<<"/* Probabilities */"<<endl;
	for(int i=0;i<functions.size();i++)
	{
		out<<functions[i]->table().size()<<endl;
		vector<Variable*> other_variables;
		Variable* curr_variable=variables[i];
		assert(curr_variable->id()==i);
		vector<Variable*> temp_variables;
		temp_variables.push_back(variables[i]);
		do_set_difference(functions[i]->variables(),temp_variables,other_variables,less_than_comparator_variable);
		
		for(int k=0;k<Variable::getDomainSize(other_variables);k++)
		{
			Variable::setAddressVIB(other_variables,k);
			for(int j=0;j<variables[i]->domain_size();j++)
			{
				variables[i]->addr_value()=j;
				int entry=Variable::getAddress(functions[i]->variables());
				out<<functions[i]->table()[entry]<<" ";
			}
			out<<endl;
		}
		out<<endl;
	}
	//Write names
	out<<endl;
	out<<"/* Names */"<<endl;
	for(int i=0;i<variables.size();i++)
	  {
	    out<<"Var"<<i<<endl;
	  }
	// Write labels
	out<<endl;
	out<<"/* Labels */"<<endl;
	for(int i=0;i<variables.size();i++)
	  {
	    for(int j=0;j<variables[i]->domain_size();j++)
	      {
		out<<"Val"<<i<<" ";
	      }
	    out<<endl;
	  }
	out<<endl;
	out.close();
}
//int main(int argc, char* argv[])
//{
//	GM gm;
//	gm.readCNF(argv[1]);
//	//gm.print();
//	vector<int> order;
//	for(int i=0;i<gm.variables.size();i++)
//		order.push_back(i);
//	JG jg(gm,1,50,order);
//	cerr<<"Propagating..\n";
//	jg.propagate();
//	cerr<<"Done\n";
//	cerr<<"Printing marginals\n";
//	for(int i=0;i<gm.variables.size();i++)
//	{
//		vector<Variable*> curr_variable;
//		curr_variable.push_back(gm.variables[i]);
//		vector<Double> marg_table;
//		for(int j=0;j<jg.nodes.size();j++)
//		{
//			if(do_set_inclusion(curr_variable,jg.nodes[j]->variables(),less_than_comparator_variable))
//			{
//				jg.nodes[j]->getMarginal(curr_variable,marg_table);
//				break;
//			}
//		}
//		cout<<"Marginal of "<<i<<" is ";
//		for(int i=0;i<marg_table.size();i++)
//		{
//			cout<<marg_table[i]<<" ";
//		}
//		cout<<endl;
//	}
//}
