#include "Function.h"
#include "GrayCode.h"
#include "myRandom.h"


void Function::dummy_multiplyMarginalize(vector<Variable*>& marg_variables_, vector<Function*>& functions, Function& f)
{
	// First iterate through the 
	vector<Variable*> variables;
	vector<Variable*> marg_variables;
	//cout<<"Computing marginals\n";
	for(int i=0;i<functions.size();i++)
	{
		do_set_union(variables,functions[i]->variables(),variables,less_than_comparator_variable);
	}
	do_set_intersection(variables,marg_variables_,marg_variables,less_than_comparator_variable);
	int num_values=Variable::getDomainSize(variables);
	f=Function();
	f.variables()=marg_variables;
	f.table()=vector<Double> (Variable::getDomainSize(marg_variables));
	for(int i=0;i<num_values;i++)
	{
		Variable::setAddress(variables,i);
		Double value(1.0);
		for(int j=0;j<functions.size();j++)
		{
			int func_entry=Variable::getAddress(functions[j]->variables());
			value*=functions[j]->table()[func_entry];
		}
		f.table()[Variable::getAddress(marg_variables)]+=value;
	}
	Double norm_const;
	for(int i=0;i<f.table().size();i++)
	{
		norm_const+=f.table()[i];
	}
	for(int i=0;i<f.table().size();i++)
	{
		f.table()[i]/=norm_const;
	}
}

void Function::multiplyAndMarginalize(vector<Variable*>& marg_variables_, vector<Function*>& functions, Function& f)
{
	// First iterate through the 
	vector<Variable*> variables;
	vector<Variable*> marg_variables;
	//cout<<"Computing marginals\n";
	for(int i=0;i<functions.size();i++)
	{
		do_set_union(variables,functions[i]->variables(),variables,less_than_comparator_variable);
	}
	do_set_intersection(variables,marg_variables_,marg_variables,less_than_comparator_variable);
	//cout<<"\tGray code\n";
	//if(variables.empty())
	//	return;
	GrayCodePM gray_code(variables,functions,&f,marg_variables);
	int num_values=Variable::getDomainSize(variables);

	//cout<<"\tMoving forward\n";

	for(int i=0;i<num_values;i++)
	{
		
		gray_code.moveForward();
	}
	Double norm_const;
	for(int i=0;i<f.table().size();i++)
	{
		norm_const+=f.table()[i];
	}
	for(int i=0;i<f.table().size();i++)
	{
		f.table()[i]/=norm_const;
	}
	//cout<<"Done\n";
}
void Function::divide(Function& f1, Function& f2, Function& f3)
{
	vector<Variable*> variables;
	vector<Function*> functions;
	functions.push_back(&f1);functions.push_back(&f2);
	for(int i=0;i<functions.size();i++)
	{
		do_set_union(variables,functions[i]->variables(),variables,less_than_comparator_variable);
	}
	
	GrayCodeD gray_code(variables,functions,&f3);
	int num_values=Variable::getDomainSize(variables);
	for(int i=0;i<num_values;i++)
	{
		gray_code.moveForward();
	}
}
void Function::product(vector<Function*>& functions, Function& f)
{
	vector<Variable*> variables;
	for(int i=0;i<functions.size();i++)
	{
		do_set_union(variables,functions[i]->variables(),variables,less_than_comparator_variable);
	}
	int num_values=Variable::getDomainSize(variables);
	GrayCodeP gray_code(variables,functions,&f);
	for(int i=0;i<num_values;i++)
	{
		gray_code.moveForward();
	}
}
void Function::dummy_product(Function& f1, Function& f2, Function& f3)
{
	do_set_union(f1.variables(),f2.variables(),f3.variables(),less_than_comparator_variable);
	int num_values=Variable::getDomainSize(f3.variables());
	f3.table()=vector<Double>(num_values);
	for(int i=0;i<num_values;i++)
	{
		Variable::setAddress(f3.variables(),i);
		int add1,add2;
		add1=Variable::getAddress(f1.variables());
		add2=Variable::getAddress(f2.variables());
		f3.table()[i]=f1.table()[add1]*f2.table()[add2];
	}
}
void Function::print(ostream& out)
{
	out<<"Variables: ";
	for(int i=0;i<variables_.size();i++)
		out<<variables_[i]->id()<<" ";
	out<<endl;
	for(int i=0;i<table_.size();i++)
		out<<table_[i]<<" ";
	out<<endl;
}
double Function::MSE(Function& function)
{
	double error=0.0;
	if(function.table().size()!=table_.size())
		return -1;
	for(int i=0;i<table_.size();i++)
	{
		error+=((function.table()[i].value()-table_[i].value())*(function.table()[i].value()-table_[i].value()));
	}
	return error;

}
double Function::MSE(Function* f1, Function* f2)
{
	double error=0.0;
	if(f1->table().size()!=f2->table().size())
		return -1;
	for(int i=0;i<f1->table().size();i++)
	{
		error+=((f1->table()[i].value()-f2->table()[i].value())*(f1->table()[i].value()-f2->table()[i].value()));
	}
	return error;
}
