#include "GM.h"
#include "JG.h"
//bool less_than_comparator_variable(const Variable* a, const Variable* b)
//{
//	if(a->id()<b->id())
//		return true;
//	return false;
//}
void Function::readCNF(istream& in, vector<Variable*>& all_variables)
{
	do{
		int curr_var;
		in>>curr_var;
		if(curr_var ==0)
			break;
		int var_id=((curr_var > 0)? (curr_var-1):(-curr_var-1));
		variables_.push_back(all_variables[var_id]);
		if(curr_var>0)
		{
			all_variables[var_id]->addr_value()=0;
		}
		else
		{
			all_variables[var_id]->addr_value()=1;
		}
		
	}while(1);
	
	sort(variables_.begin(),variables_.end(),less_than_comparator_variable);
	
	int table_size=Variable::getDomainSize(variables_);
	table_=vector<Double> (table_size);
	for(int i=0;i<table_.size();i++)
	{
		table_[i]=Double(1.0);
	}
	
	table_[Variable::getAddress(variables_)]=Double();
}

void GM::readCNF(char* infilename)
{
	ifstream xInputFile(infilename);
	char cCheck;
	char buffer[5000];
	int num_variables,num_constraints;
	while (1) 
	{
		xInputFile >> cCheck;
		if (cCheck == 'c') 
		{
			xInputFile.getline(buffer, 5000);
			continue;
		}
		else if (cCheck == 'p') 
		{
			xInputFile >> buffer;
			xInputFile >> num_variables;
			xInputFile >> num_constraints;
			break;
		}
		else 
		{
			cout << "\nError: File not in DIMACS format?\n";
			//return 0;
			exit(1);
		}
	}
	// Read the variables
	variables=vector<Variable*>(num_variables);
	for(int i=0;i<num_variables;i++)
	{
		variables[i]=new Variable(i,2);
	}
	functions=vector<Function*>(num_constraints);
	for(int i=0;i<num_constraints;i++)
		functions[i]=new Function(i);
	int i=0;
	while (1) 
	{
		if(i==num_constraints)
			break;
		if (xInputFile.eof()) 
		{
			cout << "\nError: Unexpected end of file.\n";
			//return 0;
			exit(1);
		}
		xInputFile >> cCheck;
		if (cCheck == 'c') 
		{
			xInputFile.getline(buffer, 5000);
			continue;
		}
		else xInputFile.putback(cCheck);
		functions[i]->readCNF(xInputFile,variables);
		i++;
	}
}
void GM::convertToSAT()
{

	// First map the variables between SAT and the belief network
	// For each variable we create a clause
	// Create a value for each variable value assignment
	int count=0;
	csp_to_sat_variables=vector<vector<int> > (variables.size());
	int num_con=0;
	for(int i=0;i<variables.size();i++)
	{
		for(int j=0;j<variables[i]->domain_size();j++)
		{
			sat_to_csp_variables.push_back(i);
			csp_to_sat_variables[i].push_back(count);
			count++;
		}
		/*num_con+=combination(variables[i]->num_values,2);
		if(variables[i]->num_values >1)
			num_con++;*/
	}
	for(int i=0;i<variables.size();i++)
	{
		num_con++;
		for(int j=0;j<csp_to_sat_variables[i].size();j++)
		{
			for(int k=j+1;k<csp_to_sat_variables[i].size();k++)
			{
				num_con++;
			}
		}
	}
	//// Add all evidence variables as clauses
	//for(int i=0;i<num_variables;i++)
	//	if(variables[i]->is_evidence)
	//		num_con++;
	/*for( int i=0;i<num_constraints;i++)
	{
		for(int j=0;j<constraints[i]->table.size();j++)
		{
			Variable::setAddressValues(constraints[i]->scope,j);
			if(!constraints[i]->table[constraints[i]->getEntry()])
			{
				num_con++;
			}
		}
	}*/
	//num_con+=num_constraints;
	cerr<<"Counting constraints\n";
	//Find the number of constraints that can be derived from the cpts
	for(int i=0;i<functions.size();i++)
	{
		Function* function =functions[i];
		for(int j=0;j<function->table().size();j++)
		{
			if(function->table()[j].isZero())
				num_con++;
		}
	}
	//cerr<<"Writing to file\n";
	ofstream outfile ("temp.cnf");
	outfile<<"p cnf "<<sat_to_csp_variables.size()<<" "<<num_con<<endl;

	//Now write constraints to ostream
	for(int i=0;i<variables.size();i++)
	{
		
		
		// A constraint making sure that each variable is assigned a value
		for(int j=0;j<csp_to_sat_variables[i].size();j++)
		{
			outfile<<csp_to_sat_variables[i][j]+1<<" ";
		}
		outfile<<"0 \n";
		// A constraint making sure that each variable is assigned a value
		for(int j=0;j<csp_to_sat_variables[i].size();j++)
		{
			for(int k=j+1;k<csp_to_sat_variables[i].size();k++)
			{
				outfile<<-(csp_to_sat_variables[i][j]+1)<<" "<<-(csp_to_sat_variables[i][k]+1)<<" 0 "<<endl;
			}
		}

	}
	for(int i=0;i<functions.size();i++)
	{
		Function* function = functions[i];
		
		for(int j=0;j<function->table().size();j++)
		{
			Variable::setAddress(function->variables(),j);
			if(function->table()[j].isZero())
			{
			
				for(int k=0;k<function->variables().size();k++)
				{
					/*if(function->scope[k]->addr_value==0)
					{
						outfile<<(function->scope[k]->id+1)<<" ";
					}
					else
					{
						outfile<<-(function->scope[k]->id+1)<<" ";
					}*/
					outfile<<-(csp_to_sat_variables[function->variables()[k]->id()][function->variables()[k]->addr_value()]+1)<<" ";
				}
				outfile<<"0\n";
			}
		}
	}

	// Write evidence
	/*for(int i=0;i<num_variables;i++)
		if(variables[i]->is_evidence)
		{
			assert(variables[i]->value!=INVALID_VALUE);
			outfile<<csp_to_sat_variables[i][variables[i]->value]+1<<" 0\n";
		}*/
	/*for( int i=0;i<num_constraints;i++)
	{
		for(int j=0;j<constraints[i]->table.size();j++)
		{
			Variable::setAddressValues(constraints[i]->scope,j);
			if(!constraints[i]->table[constraints[i]->getEntry()])
			{
				for(int k=0;k<constraints[i]->scope_size;k++)
				{
					int var=constraints[i]->scope[k]->id;
					int val=constraints[i]->scope[k]->addr_value;
					outfile<<-(csp_to_sat_variables[var][val]+1)<<" ";
				}
				outfile<<" 0\n";
			}
		}
	}*/
	outfile.close();

}

void GM::readVIB(char* bayes_file)
{
	cout<<"Reading the Bayesian Network..."<<flush;
	//Open the file
	ifstream infile(bayes_file);
	if(!infile.good())
	{
		cerr<<"Empty Bayesnet file\n";
		return;
	}
	int num_variables;
	// Read number of variables
	infile>>num_variables;
	//Read the values of each variable
	variables=vector<Variable*>(num_variables);
	
	//Read variable information
	for(int i=0;i<num_variables;i++)
	{
		int num_values;
		infile>>num_values;
		vector<int> domain;
		for(int j=0;j<num_values;j++)
		{
			domain.push_back(j);
		}
		variables[i]=new Variable(i,domain);
	}
	
	int num_cpts;
	//Read #conditional probability tables
	infile>>num_cpts;
	
	functions=vector<Function*> (num_cpts);

	// Currently we only allow num cpts = numvariables
	assert(num_cpts==num_variables);
	int max_parents=0;
	//Read the cpts
	for(int i=0;i<num_cpts;i++)
	{
		functions[i]=new Function(i);
		vector<Variable*> func_variables;
		vector<Variable*> sorted_func_variables;
		int num_variables_infunc;
		
		//Read the variables in the function
		infile>>num_variables_infunc;
		int marg_variable;
		infile>>marg_variable;
		func_variables.push_back(variables[marg_variable]);
		for(int j=1;j<num_variables_infunc;j++)
		{
			int temp;
			infile>>temp;
			func_variables.push_back(variables[temp]);
		}
		if(max_parents < func_variables.size())
			max_parents=func_variables.size();
		// Currently in vib format we allow specification of functions in order
		assert(marg_variable==i);

		//func_variables.push_back(variables[marg_variable]);
		//Read the table
		sorted_func_variables=func_variables;
		sort(sorted_func_variables.begin(),sorted_func_variables.end(),
			less_than_comparator_variable);
		//sort(func_variables.begin(),func_variables.end(),less_than_operator);
		int num_values=Variable::getDomainSize(sorted_func_variables);
		//int marg_num_values=Variable::getNumValues(marg_variables);
		functions[i]->variables()=sorted_func_variables;
		functions[i]->table()=vector<Double> (num_values);
		
		for(int j=0;j<num_values;j++)
		{
			double value;
			infile>>value;
			Variable::setAddress(func_variables,j);
			int entry=Variable::getAddress(sorted_func_variables);
			if(value > DBL_MIN)
			{
				
				functions[i]->table()[entry]=Double(value);
			}
		}	
	}

	// Construct the graph of the bayesian network
	cout<<"Max parents = "<<max_parents+1<<endl;
}
void GM::readEvidence1(char* evidencefile)
{
	mult_factor=Double((double)1.0);
	int num_evidence;
	ifstream evidfile(evidencefile);
	evidfile>>num_evidence;
	for(int i=0;i<num_evidence;i++)
	{
		int var,val;
		evidfile>>var>>val;
		assert(var < variables.size());
		assert(val < variables[var]->domain_size());
		Function *func=new Function();
		func->variables().push_back(variables[var]);
		func->table()=vector<Double>(variables[var]->domain_size());
		func->table()[val]=Double((double)1.0);
		functions.push_back(func);
	}
}

void GM::readEvidence(char* evidencefile)
{
	mult_factor=Double((double)1.0);
	vector<Variable*> evidence_variables;

	int num_evidence;
	ifstream evidfile(evidencefile);
	evidfile>>num_evidence;
	for(int i=0;i<num_evidence;i++)
	{
		int var,val;
		evidfile>>var>>val;
		evidence_variables.push_back(variables[var]);
		cerr<<var<<" "<<val<<endl;
		variables[var]->value()=val;
		variables[var]->addr_value()=val;
	}
	sort(evidence_variables.begin(),evidence_variables.end(),less_than_comparator_variable);
	cerr<<"num-evidence = "<<evidence_variables.size()<<endl;
	vector<int> variable_mapping(variables.size());
	vector<bool> is_evidence(variables.size());
	for(int i=0;i<evidence_variables.size();i++)
		is_evidence[i]=true;

	int var_count=0;
	for(int i=0;i<variable_mapping.size();i++)
	{
		if(!is_evidence[i])
		{
			
			variable_mapping[i]=var_count;
			var_count++;
		}
		else
		{
			variable_mapping[i]=INVALID_VALUE;
		}
	}

	vector<bool> deleted_functions(functions.size());
	//update functions
	for(int i=0;i<functions.size();i++)
	{
		vector<Variable*> func_evid_vars;
		deleted_functions[i]=false;
		do_set_intersection(functions[i]->variables(),evidence_variables,func_evid_vars,less_than_comparator_variable);
		if(func_evid_vars.empty())
			continue;
		// All variables are evidence variables
		if(func_evid_vars.size()==functions[i]->variables().size())
		{
			int address=Variable::getAddress(functions[i]->variables());
			mult_factor*=functions[i]->table()[address];
			
			deleted_functions[i]=true;
			continue;
		}
		vector<Variable*> other_variables;
		do_set_difference(functions[i]->variables(),func_evid_vars,other_variables,less_than_comparator_variable);
		int other_num_values=Variable::getDomainSize(other_variables);
		vector<Double> new_table(other_num_values);
		for(int j=0;j<other_num_values;j++)
		{
			Variable::setAddress(other_variables,j);
			int entry=Variable::getAddress(functions[i]->variables());
			new_table[j]=functions[i]->table()[entry];
		}
		functions[i]->table()=new_table;
		functions[i]->variables()=other_variables;
	}
	cerr<<"Mult-factor = "<<mult_factor<<endl;
	//update functions;
	vector<Function*> new_functions;
	int count=0;
	for(int i=0;i<functions.size();i++)
	{
		if(deleted_functions[i])
		{
			delete(functions[i]);
		}
		else
		{
			functions[i]->id()=count++;
			new_functions.push_back(functions[i]);
		}
	}
	functions=new_functions;
	//update variables
	vector<Variable*> new_variables;
	for(int i=0;i<variables.size();i++)
	{
		if(is_evidence[i])
			delete(variables[i]);
		else
		{
			variables[i]->id()=variable_mapping[i];
			new_variables.push_back(variables[i]);
		}
	}
	variables=new_variables;
	//for(int i=0;i<variables.size();i++)
	//	cout<<variables[i]->id()<<" ";
	//cout<<endl;
}

//int main(int argc, char* argv[])
//{
//	GM gm;
//	gm.readCNF(argv[1]);
//	//gm.print();
//	vector<int> order;
//	for(int i=0;i<gm.variables.size();i++)
//		order.push_back(i);
//	JG jg(gm,1,50,order);
//	cerr<<"Propagating..\n";
//	jg.propagate();
//	cerr<<"Done\n";
//	cerr<<"Printing marginals\n";
//	for(int i=0;i<gm.variables.size();i++)
//	{
//		vector<Variable*> curr_variable;
//		curr_variable.push_back(gm.variables[i]);
//		vector<Double> marg_table;
//		for(int j=0;j<jg.nodes.size();j++)
//		{
//			if(do_set_inclusion(curr_variable,jg.nodes[j]->variables(),less_than_comparator_variable))
//			{
//				jg.nodes[j]->getMarginal(curr_variable,marg_table);
//				break;
//			}
//		}
//		cout<<"Marginal of "<<i<<" is ";
//		for(int i=0;i<marg_table.size();i++)
//		{
//			cout<<marg_table[i]<<" ";
//		}
//		cout<<endl;
//	}
//}