#include "JG.h"
JGNode* JG::addNode(JG_TYPE type)
{
	switch(type)
	{
	case SS:
		return new JGNodeSS();
		break;
	case LS:
		return new JGNodeLS();
		break;
	case HSS:
		return new JGNodeHSS();
		break;
	default:
		return new JGNodeSS();
	}
}

JGEdge* JG::addEdge(JGNode* s1, JGNode* s2, JG_TYPE type)
{
	switch (type)
	{
	case SS:
		return new JGEdgeSS(dynamic_cast<JGNodeSS*> (s1),dynamic_cast<JGNodeSS*> (s2));
		break;
	case LS:
		return new JGEdgeLS(dynamic_cast<JGNodeLS*> (s1),dynamic_cast<JGNodeLS*> (s2));
		break;
	case HSS:
		return new JGEdgeHSS(dynamic_cast<JGNodeHSS*> (s1),dynamic_cast<JGNodeHSS*> (s2));
		break;
	default:
		return new JGEdgeSS(dynamic_cast<JGNodeSS*> (s1),dynamic_cast<JGNodeSS*> (s2));
		break;
	}
}
void JGNodeSS::addFunction(Function& function)
{
	this->functions.push_back(&function);
}
void JGNodeSS::compileAllFunctions(vector<Function*>& all_functions)
{
	all_functions=functions;
	for(int i=0;i<edges_.size();i++)
	{
		if(edges_[i]->node1()==this)
		{
			if(edges_[i]->message2().variables().empty())
				continue;
			all_functions.push_back(&(edges_[i]->message2()));
		}
		else
		{
			if(edges_[i]->message1().variables().empty())
				continue;
			all_functions.push_back(&(edges_[i]->message1()));
		}
	}
}
void JGNodeSS::getCF(vector<Variable*>& cond_variables_, Variable* marg_variable, CPT& cf)
{
	vector<Function*> all_functions;
	compileAllFunctions(all_functions);
	vector<Variable*> all_variables,wanted_variables,marg_vars,cond_variables;
	marg_vars.push_back(marg_variable);
	for(int i=0;i<all_functions.size();i++)
	{
		do_set_union(all_variables,all_functions[i]->variables(),all_variables,less_than_comparator_variable);
	}
	
	wanted_variables=cond_variables_;
	wanted_variables.push_back(marg_variable);
	sort(wanted_variables.begin(),wanted_variables.end(),less_than_comparator_variable);

	do_set_intersection(wanted_variables,all_variables,all_variables,less_than_comparator_variable);
	do_set_intersection(all_variables,cond_variables_,cond_variables,less_than_comparator_variable);
	do_set_intersection(all_variables,marg_vars,marg_vars,less_than_comparator_variable);
	assert(!marg_vars.empty());

	Function func;
	Function::dummy_multiplyMarginalize(all_variables,all_functions,func);
	
	cf=CPT(func,marg_variable,cond_variables);
	
}

void JGNodeSS::getMarginal(vector<Variable*>& marg_variables, Function& function)
{
	vector<Function*> all_functions;
	all_functions=functions;
	for(int i=0;i<edges_.size();i++)
	{
		if(edges_[i]->node1()==this)
		{
			all_functions.push_back(&(edges_[i]->message2()));
		}
		else
		{
			all_functions.push_back(&(edges_[i]->message1()));
		}
	}
	Function::multiplyAndMarginalize(marg_variables,all_functions,function);
}

void JGNodeHSS::addFunction(Function& function)
{
	functions.push_back(new HashFunction(function));
}
void JGNodeHSS::getCF(vector<Variable*>& cond_variables,Variable* marg_variable, CPT& cf)
{
	vector<HashFunction*> all_functions;
	all_functions=functions;
	for(int i=0;i<edges_.size();i++)
	{
		if(edges_[i]->node1()==this)
		{
			HashFunction* h;
			Function* f;
			f=&(edges_[i]->message2());
			h=dynamic_cast<HashFunction*>(f);
			all_functions.push_back(h);
		}
		else
		{
			HashFunction* h;
			Function* f;
			f=&(edges_[i]->message1());
			h=dynamic_cast<HashFunction*>(f);
			all_functions.push_back(h);
			
		}
	}
	vector<Variable*> all_variables;
	all_variables=cond_variables;
	all_variables.push_back(marg_variable);
	sort(all_variables.begin(),all_variables.end(),less_than_comparator_variable);

	HashFunction temp;
	HashFunction::multiplyAndMarginalize(all_variables,all_functions,&temp);
	cf.variables()=all_variables;
	cf.cond_variables()=cond_variables;
	cf.setMargVariable(marg_variable);
	cf.table()=vector<Double> (Variable::getDomainSize(all_variables));
	for(int i=0;i<cf.table().size();i++)
		cf.table()[i]=temp.stable().get(i);
	//Normalize the conditional function
	int multiplier=1;
	for(int i=0;i<all_variables.size();i++)
	{
		if(all_variables[i]==marg_variable)
			break;
		multiplier*=all_variables[i]->domain_size();
	}
	int cond_num_values=Variable::getDomainSize(cond_variables);
	for(int i=0;i<cond_num_values;i++)
	{
		Double norm_const;
		for(int j=0;j<marg_variable->domain_size();j++)
		{
			int address=i+multiplier*j;
			norm_const+=cf.table()[address];
		}
		for(int j=0;j<marg_variable->domain_size();j++)
		{
			int address=i+multiplier*j;
			cf.table()[address]/=norm_const;
		}
	}
}

void JGNodeHSS::getMarginal(vector<Variable*>& marg_variables, Function& function)
{
	vector<HashFunction*> all_functions;
	all_functions=functions;
	for(int i=0;i<edges_.size();i++)
	{
		if(edges_[i]->node1()==this)
		{
			HashFunction* h;
			Function* f;
			f=&(edges_[i]->message2());
			h=dynamic_cast<HashFunction*>(f);
			all_functions.push_back(h);
		}
		else
		{
			HashFunction* h;
			Function* f;
			f=&(edges_[i]->message1());
			h=dynamic_cast<HashFunction*>(f);
			all_functions.push_back(h);
			
		}
	}
	vector<Variable*> all_variables;
	all_variables=marg_variables;
	sort(all_variables.begin(),all_variables.end(),less_than_comparator_variable);

	HashFunction temp;
	HashFunction::multiplyAndMarginalize(all_variables,all_functions,&temp);
	function.variables()=all_variables;
	function.table()=vector<Double> (Variable::getDomainSize(all_variables));
	for(int i=0;i<function.table().size();i++)
		function.table()[i]=temp.stable().get(i);
}
void JGNodeLS::addFunction(Function& function_)
{
	Function new_function;
	vector<Function*> all_functions;
	all_functions.push_back(&function);
	all_functions.push_back(&function_);
	Function::product(all_functions,new_function);
	function=new_function;
}
void JGNodeLS::getCF(vector<Variable*>& cond_variables, Variable* marg_variable, CPT& cf)
{
	vector<Function*> all_functions;
	all_functions.push_back(&function);
	vector<Variable*> all_variables;
	all_variables=cond_variables;
	all_variables.push_back(marg_variable);
	sort(all_variables.begin(),all_variables.end(),less_than_comparator_variable);

	Function::multiplyAndMarginalize(all_variables,all_functions,cf);
	cf.cond_variables()=cond_variables;
	cf.setMargVariable(marg_variable);
	//Normalize the conditional function
	int multiplier=1;
	for(int i=0;i<all_variables.size();i++)
	{
		if(all_variables[i]==marg_variable)
			break;
		multiplier*=all_variables[i]->domain_size();
	}
	int cond_num_values=Variable::getDomainSize(cond_variables);
	for(int i=0;i<cond_num_values;i++)
	{
		Double norm_const;
		for(int j=0;j<marg_variable->domain_size();j++)
		{
			int address=i+multiplier*j;
			norm_const+=cf.table()[address];
		}
		for(int j=0;j<marg_variable->domain_size();j++)
		{
			int address=i+multiplier*j;
			cf.table()[address]/=norm_const;
		}
	}
}
void JGNodeLS::getMarginal(vector<Variable*>& marg_variables, Function& function_)
{
	vector<Function*> all_functions;
	all_functions.push_back(&function);
	Function::multiplyAndMarginalize(marg_variables,all_functions,function_);
}


JGEdgeSS::JGEdgeSS(JGNodeSS* ss_node1__, JGNodeSS* ss_node2__): JGEdge(),ss_node1_(ss_node1__),ss_node2_(ss_node2__)
{
	node1_=ss_node1_;
	node2_=ss_node2_;
}
void JGEdgeSS::sendMessage1to2()
{
	vector<Function*> all_functions;
	all_functions=ss_node1_->functions;
	for(int i=0;i<ss_node1_->edges().size();i++)
	{
		JGEdge* curr_edge=ss_node1_->edges()[i];
		if(curr_edge==this)
		{
			//cout<<"Here .. ";
			continue;
		}
		if(curr_edge->node1()->id()==ss_node1_->id())
		{
			if(curr_edge->message2().table().empty())
				continue;
			all_functions.push_back(&curr_edge->message2());
		}
		else
		{
			if(curr_edge->message1().table().empty())
				continue;
			all_functions.push_back(&curr_edge->message1());
		}
	}
	//cout<<"Calling multiply and marginalize ";
	//cout<<ss_node1_->id()<<" to "<<ss_node2_->id()<<endl;
	//Function::multiplyAndMarginalize(this->variables(),all_functions,this->message1());
	Function::dummy_multiplyMarginalize(this->variables(),all_functions,this->message1());
}

void JGEdgeSS::sendMessage2to1()
{
	vector<Function*> all_functions;
	all_functions=ss_node2_->functions;
	for(int i=0;i<ss_node2_->edges().size();i++)
	{
		JGEdge* curr_edge=ss_node2_->edges()[i];
		if(curr_edge==this)
		{
			//cout<<"Here .. ";
			continue;
		}
		if(curr_edge->node1()->id()==ss_node1_->id())
		{
			if(curr_edge->message2().table().empty())
				continue;
			all_functions.push_back(&curr_edge->message2());
		}
		else
		{
			if(curr_edge->message1().table().empty())
				continue;
			all_functions.push_back(&curr_edge->message1());
		}
	}
	//cout<<"Calling multiply and marginalize ";
	//cout<<ss_node2_->id()<<" to "<<ss_node1_->id()<<endl;
	//Function::multiplyAndMarginalize(this->variables(),all_functions,this->message2());
	Function::dummy_multiplyMarginalize(this->variables(),all_functions,this->message2());
}

JGEdgeHSS::JGEdgeHSS(JGNodeHSS* hss_node1__, JGNodeHSS* hss_node2__): JGEdge(),hss_node1_(hss_node1__),hss_node2_(hss_node2__)
{
	if(node1_to_node2_message_!=NULL)
	{
		delete(node1_to_node2_message_);
	}
	if(node2_to_node1_message_!=NULL)
	{
		delete(node2_to_node1_message_);
	}
	node1_to_node2_message_=new HashFunction();
	node2_to_node1_message_=new HashFunction();
	node1_=hss_node1_;
	node2_=hss_node2_;
}
void JGEdgeHSS::sendMessage1to2()
{
	vector<HashFunction*> all_functions;
	all_functions=hss_node1_->functions;
	for(int i=0;i<hss_node1_->edges().size();i++)
	{
		JGEdge* curr_edge=hss_node1_->edges()[i];
		if(curr_edge==this)
			continue;
		if(curr_edge->node1()==hss_node1_)
		{
			if(!curr_edge->message2().variables().empty())
			{
				all_functions.push_back(dynamic_cast<HashFunction*>(&curr_edge->message2()));
			}
		}
		else
		{
			if(!curr_edge->message1().variables().empty())
			{
				all_functions.push_back(dynamic_cast<HashFunction*>(&curr_edge->message1()));
			}
		}
	}
	//cout<<"Calling multiply and marginalize hss ";
	//cout<<hss_node1_->id()<<" to "<<hss_node2_->id()<<endl;
	HashFunction::multiplyAndMarginalize(this->variables(),all_functions,dynamic_cast<HashFunction*>(node1_to_node2_message_));
}
void JGEdgeHSS::sendMessage2to1()
{
	vector<HashFunction*> all_functions;
	all_functions=hss_node2_->functions;
	for(int i=0;i<hss_node2_->edges().size();i++)
	{
		JGEdge* curr_edge=hss_node2_->edges()[i];
		if(curr_edge==this)
			continue;
		if(curr_edge->node1()==hss_node2_)
		{
			if(!curr_edge->message2().variables().empty())
				all_functions.push_back(dynamic_cast<HashFunction*>(&curr_edge->message2()));
		}
		else
		{
			if(!curr_edge->message2().variables().empty())
				all_functions.push_back(dynamic_cast<HashFunction*>(&curr_edge->message1()));
		}
	}
	//cout<<"Calling multiply and marginalize ";
	//cout<<hss_node2_->id()<<" to "<<hss_node1_->id()<<endl;
	HashFunction::multiplyAndMarginalize(this->variables(),all_functions,dynamic_cast<HashFunction*>(node2_to_node1_message_));
}
JGEdgeLS::JGEdgeLS(JGNodeLS* ls_node1__, JGNodeLS* ls_node2__): JGEdge(), ls_node1_(ls_node1__),ls_node2_(ls_node2__)
{
	node1_=ls_node1_;
	node2_=ls_node2_;
}
void JGEdgeLS::sendMessage1to2()
{
	vector<Function*> all_functions;
	Function f1,f2,f3,f4;
	Function::divide(ls_node1_->function,message2(),f1);
	all_functions.push_back(&f1);
	f2=this->message1();
	Function::multiplyAndMarginalize(this->variables(),all_functions,this->message1());
	all_functions=vector<Function*>();
	Function::divide(ls_node2_->function,f2,f3);
	all_functions.push_back(&f3);
	all_functions.push_back(&ls_node2_->function);
	Function::multiplyAndMarginalize(ls_node2_->variables(),all_functions,f4);
	ls_node2_->function=f4;
}

void JGEdgeLS::sendMessage2to1()
{
	vector<Function*> all_functions;
	Function f1,f2,f3,f4;
	Function::divide(ls_node2_->function,message1(),f1);
	all_functions.push_back(&f1);
	f2=this->message2();
	Function::multiplyAndMarginalize(this->variables(),all_functions,this->message2());
	all_functions=vector<Function*>();
	Function::divide(ls_node1_->function,f2,f3);
	all_functions.push_back(&f3);
	all_functions.push_back(&ls_node1_->function);
	Function::multiplyAndMarginalize(ls_node1_->variables(),all_functions,f4);
	ls_node1_->function=f4;
}
//void JGEdge::printMessages()
//{
//	cout<<"Message1 : ";
//	for(int i=0;i<message1().size();i++)
//		cout<<message1()[i]<<" ";
//	cout<<" message 2: ";
//	for(int i=0;i<message2().size();i++)
//		cout<<message2()[i]<<" ";
//	cout<<endl;
//
//}

struct JGHelperFunction
{
	vector<Variable*> variables;
	int parent_id;
	Function* f;
	JGHelperFunction(vector<Variable*>& variables_,int id_,Function* func_=NULL):variables(variables_),parent_id(id_), f(func_)
	{
	}
};

JG::JG(GM& gm, int i_bound_, int num_iterations__, vector<int>& order, JG_TYPE type): num_iterations_(num_iterations__)
{
	vector<int> mapped_order(order.size());
	for(int i=0;i<order.size();i++)
		mapped_order[order[i]]=i;
	assert(order.size()==gm.variables.size());
	//cerr<<"Creating buckets\n";
	// Create buckets
	vector<vector<JGHelperFunction> > buckets (order.size());
	for(int i=0;i<gm.functions.size();i++)
	{
		int min_size=(int)gm.variables.size();
		for(int j=0;j<gm.functions[i]->variables().size();j++)
		{
			if(mapped_order[gm.functions[i]->variables()[j]->id()] < min_size)
			{
				min_size=mapped_order[gm.functions[i]->variables()[j]->id()];
			}		
		}
		if(min_size == (int)gm.variables.size())
		{
			for(int j=0;j<gm.functions[i]->variables().size();j++)
			{
				cout<<gm.functions[i]->variables()[j]->id()<<endl;
			}

		}
		//cout<<min_size<<endl;
		buckets[min_size].push_back(JGHelperFunction(gm.functions[i]->variables(),INVALID_VALUE,gm.functions[i]));
		//cout<<"\tAdded\n";
	}
	//cout<<"buckets created\n";

	//Update i_bound
	int max_func_size=0;
	
	for(int i=0;i<gm.functions.size();i++)
	{
		int domain_size=0;
		domain_size=Variable::getDomainSize(gm.functions[i]->variables());
		if(max_func_size < domain_size)
		{
			max_func_size=domain_size;
		}
	}
	int avg_domain_size=0;
	for(int i=0;i<gm.variables.size();i++)
	{
		avg_domain_size+=gm.variables[i]->domain_size();
	}
	avg_domain_size/=(int)gm.variables.size();
	//cerr<<avg_domain_size<<" "<<max_func_size<<" ";
	int i_bound=(max_func_size)*(int)pow((double)avg_domain_size,(double)(i_bound_-1));
	max_cluster_size=i_bound;
	//cerr<<"i-bound = "<<i_bound<<endl;
	int old_nodes=0;
	int id=0;
	// Run schematic mini-buckets
	for(int i=0;i<order.size();i++)
	{
		//cerr<<i<<endl;
		// Form mini-buckets by using the first fill rule
		for(int j=0;j<buckets[i].size();j++)
		{
			bool found=false;
			int loc=INVALID_VALUE;
			for(int k=old_nodes;k<nodes.size();k++)
			{
				vector<Variable*> temp_variables;
				do_set_union(buckets[i][j].variables,nodes[k]->variables(),temp_variables,less_than_comparator_variable);
				int domain_size=Variable::getDomainSize(temp_variables);
				if((int)domain_size <= i_bound)
				{
					found=true;
					loc=k;
					break;
				}
			}
			// If not found create a mini-bucket
			if(!found)
			{
				
				nodes.push_back(addNode(type));
				nodes[nodes.size()-1]->id()=id;
				id++;
				nodes[nodes.size()-1]->variables()=buckets[i][j].variables;
				if(buckets[i][j].parent_id !=INVALID_VALUE)
				{
					//cout<<"F1\n";
					//nodes[nodes.size()-1]->neighbors().push_back(nodes[buckets[i][j].parent_id]);
					//nodes[buckets[i][j].parent_id]->neighbors().push_back(nodes[nodes.size()-1]);
					JGEdge* edge=addEdge(nodes[nodes.size()-1],nodes[buckets[i][j].parent_id],type);
					do_set_intersection(nodes[nodes.size()-1]->variables(),nodes[buckets[i][j].parent_id]->variables(),edge->variables(),less_than_comparator_variable);
					//edge->setNode1(nodes[nodes.size()-1]);
					//edge->setNode2(nodes[buckets[i][j].parent_id]);
					do_set_intersection(nodes[nodes.size()-1]->variables(),nodes[buckets[i][j].parent_id]->variables(),edge->variables(),less_than_comparator_variable);
					nodes[nodes.size()-1]->edges().push_back(edge);
					nodes[buckets[i][j].parent_id]->edges().push_back(edge);
					
				}
				else
				{
					//cout<<"F2\n";
					nodes[nodes.size()-1]->addFunction(*buckets[i][j].f);
				}
			}
			else
			{

				do_set_union(buckets[i][j].variables,nodes[loc]->variables(),nodes[loc]->variables(),less_than_comparator_variable);
				if(buckets[i][j].parent_id !=INVALID_VALUE)
				{
					//cout<<"F3\n";
					/*nodes[loc]->neighbors().push_back(nodes[buckets[i][j].parent_id]);
					nodes[buckets[i][j].parent_id]->neighbors().push_back(nodes[loc]);*/
					JGEdge* edge=addEdge(nodes[loc],nodes[buckets[i][j].parent_id],type);
					do_set_intersection(nodes[loc]->variables(),nodes[buckets[i][j].parent_id]->variables(),edge->variables(),less_than_comparator_variable);
					//edge->setNode1(nodes[loc]);
					//edge->setNode2(nodes[buckets[i][j].parent_id]);
					do_set_intersection(nodes[loc]->variables(),nodes[buckets[i][j].parent_id]->variables(),edge->variables(),less_than_comparator_variable);
					nodes[loc]->edges().push_back(edge);
					nodes[buckets[i][j].parent_id]->edges().push_back(edge);

					
				}
				else
				{
					//cout<<"F4\n";
					nodes[loc]->addFunction(*buckets[i][j].f);
				}

			}
		}
		vector<Variable*> curr_variable;
		curr_variable.push_back(gm.variables[order[i]]);	
		// Connect the newly created nodes to each other
		for(int j=old_nodes;j<nodes.size()-1;j++)
		{
			//nodes[j]->neighbors().push_back(nodes[j+1]);
			//nodes[j+1]->neighbors().push_back(nodes[j]);
			JGEdge* edge=addEdge(nodes[j],nodes[j+1],type);
			//edge->setNode1(nodes[j]);
			//edge->setNode2(nodes[j+1]);
			edge->variables()=curr_variable;
			nodes[j]->edges().push_back(edge);
			nodes[j+1]->edges().push_back(edge);
		}

		if(i<order.size()-1)
		{
			for(int j=old_nodes;j<nodes.size();j++)
			{
				vector<Variable*> temp_variables;
				do_set_difference(nodes[j]->variables(),curr_variable,temp_variables,less_than_comparator_variable);
				// Put the node in the appropriate bucket
				int min_size=(int)gm.variables.size();
				for(int k=0;k<temp_variables.size();k++)
				{
					if(min_size > mapped_order[temp_variables[k]->id()])
					{
						min_size=mapped_order[temp_variables[k]->id()];
					}
				}
				//assert(min_size<(int)gm.variables.size());
				if(min_size >=(int) gm.variables.size())
					continue;
				buckets[min_size].push_back(JGHelperFunction(temp_variables,j));
			}
		}
		old_nodes=(int)nodes.size();
	}
	
	//print();
	// Minimize edges
	//minimize();
	// update node ids
	/*for(int i=0;i<nodes.size();i++)
	{
		nodes[i]->id()=i;
	}*/
	//print();
	//// Set node and edge sizes
	//for(int i=0;i<nodes.size();i++)
	//{
	//	// Find domain size
	//	int domain_size=Variable::getDomainSize(nodes[i]->variables());
	//	nodes[i]->table()=vector<Double>(domain_size);
	//	for(int j=0;j<domain_size;j++)
	//	{
	//		nodes[i]->table()[j]=Double(1.0);
	//	}
	//	for(int j=0;j<nodes[i]->edges().size();j++)
	//	{
	//		int domain_size=Variable::getDomainSize(nodes[i]->edges()[j]->variables());
	//		nodes[i]->edges()[j]->message1()=vector<Double> (domain_size);
	//		nodes[i]->edges()[j]->message2()=vector<Double> (domain_size);
	//		for(int k=0;k<domain_size;k++)
	//		{
	//			nodes[i]->edges()[j]->message1()[k]=Double(1.0);
	//			nodes[i]->edges()[j]->message2()[k]=Double(1.0);
	//		}
	//	}
	//}
	//cerr<<"Now putting functions\n";
	////reduce();

}

bool JG::propagate()
{
	for(int i=0;i<num_iterations_;i++)
	{
		cerr<<"Iteration "<<i<<"Done\n";
		for(int j=0;j<nodes.size();j++)
		{
			for(int k=0;k<nodes[j]->edges().size();k++)
			{
				if(nodes[j]->edges()[k]->node1()->id()==nodes[j]->id())
				{
					nodes[j]->edges()[k]->sendMessage1to2();
					/*for(int a=0;a<nodes[j]->edges()[k]->node2()->table().size();a++)
					{
					cout<<nodes[j]->edges()[k]->node2()->table()[a]<<" ";
					}
					cout<<endl;*/
				}
				else
				{
					assert(nodes[j]->edges()[k]->node2()->id()==nodes[j]->id());
					nodes[j]->edges()[k]->sendMessage2to1();
				}
			}

		}
	}
	for(int i=0;i<nodes.size();i++)
	{
		/*for(int j=0;j<nodes[i]->variables().size();j++)
			cout<<nodes[i]->variables()[j]->id()<<" ";
		cout<<"---";*/
		nodes[i]->updateVariables();
		/*for(int j=0;j<nodes[i]->variables().size();j++)
			cout<<nodes[i]->variables()[j]->id()<<" ";
		cout<<endl;*/
	}
}
void JG::print(ostream& out)
{
	out<<"num nodes = "<<nodes.size()<<endl;
	for(int i=0;i<nodes.size();i++)
	{
		for(int j=0;j<nodes[i]->variables().size();j++)
		{
			out<<nodes[i]->variables()[j]->id()<<" ";
		}
		out<<endl;
	}
}

void JG::printGraph(ostream& out)
{
	out<<"num nodes = "<<nodes.size()<<endl;
	for(int i=0;i<nodes.size();i++)
	{
		for(int j=0;j<nodes[i]->variables().size();j++)
		{
			out<<nodes[i]->variables()[j]->id()<<" ";
		}
		out<<endl;
		JGNodeSS* ss_node;
		ss_node=dynamic_cast<JGNodeSS*>(nodes[i]);
		if(ss_node!=NULL)
		{
			for(int j=0;j<ss_node->functions.size();j++)
			{
				out<<"\t";
				ss_node->functions[j]->print(out);
			}
		}
		
	}
	
	for(int i=0;i<nodes.size();i++)
	{
		for(int j=0;j<nodes[i]->edges().size();j++)
		{
			JGEdge* edge=nodes[i]->edges()[j];
			out<<edge->node1()->id()<<" "<<edge->node2()->id()<<"  vars = ";
			for(int k=0;k<edge->variables().size();k++)
				out<<edge->variables()[k]->id()<<" ";
			out<<endl;
			
		}
	}
}