#include "GM.h"
#include "JG.h"
#include "IJGP_Sampler.h"
#include "SampleSearch.h"
#include <cmath>
#include <cstdlib>
#include <getopt.h>

using namespace std;

typedef enum 
{
MINFILL,TOPOLOGICAL,MINDEGREE
}ORDERING;



static struct option long_options[]=
    {
      {"ordering",required_argument,NULL,'o'},
      {"i-bound",required_argument,NULL,'i'},
	  {"p-bound",required_argument,NULL,'p'},
	  {"w-bound",required_argument,NULL,'w'},
	  {"num-samples",required_argument,NULL,'n'},
	  {"time-limit",required_argument,NULL,'t'},
	  {"num-iterations",required_argument,NULL,'j'},
	  {"input",required_argument,NULL,'f'},
      {"help", no_argument,NULL,'h'},
      
    };

static const char OPTION_STRING[]="o:i:p:w:n:t:j:f:h";

void print_help()
{
  cout<<"ijgpss [options] -f inputfile  \n";
  cout<<" --ordering <minfill,topological,mindegree> : What ordering to use\n";
  cout<<" --i-bound <int> : An integer value for the i-bound\n";
  cout<<" --p-bound <int> : An integer value for the p-bound\n";
  cout<<" --w-bound <int> : An integer value for the w-cutset bound\n";
  cout<<" --num-samples <int> : The maximum number of samples to be drawn\n";
  cout<<" --time-limt <int> : The maximum time in seconds for which the algorithm is run\n";
  cout<<" --num-iterations<int> : The maximum number of iterations for which IJGP is run\n";
  cout<<" --help : Print this help\n";
  

}

int main(int argc, char *argv[])
{
  
	ORDERING ordering=TOPOLOGICAL;
	int i_bound=3,p_bound=0,w_bound=0,num_samples=-1,time_limit=100,num_iterations=10;
	GM gm;
  if(argc==1)
    {
      print_help();
      return 1;
    }
  while(1)
    {

      int option_index =0;
      int opt = getopt_long(argc,argv,OPTION_STRING,long_options,&option_index);
      // Detect end of options
      if (opt == -1)
        break;
      switch(opt)
        {
        case 'o':
			char temp[100];
			sscanf(optarg,"%s",&temp);
			break;
		case 'i':
			sscanf(optarg,"%ld",&i_bound);
			break;
		case 'p':
			sscanf(optarg,"%ld",&p_bound);
			break;
		case 'n':
			sscanf(optarg,"%ld",&num_samples);
			break;
        case 't':
          sscanf(optarg,"%ld",&time_limit);
          break;
        case 'w':
          sscanf(optarg,"%ld",&w_bound);
          break;
		 case 'j':
          sscanf(optarg,"%ld",&num_iterations);
          break;
        
        case 'f':
			gm.readCNF(optarg);          
          break;
        
        case 'h':
          print_help();
          return 1;
		default:
			print_help();
          return 1;
        }
    }

	vector<int> order(gm.variables.size());
	cerr<<"num variables ="<<gm.variables.size()<<endl;
	for(int i=0;i<gm.variables.size();i++)
	{
		order[i]=i;
	}
	JG jg(gm,i_bound,num_iterations,order,SS);
	//jg.printGraph(cerr);
	cerr<<"Join graph constructed\n";
	time_t start_time,end_time;
	start_time=time(NULL);
	jg.propagate();
	end_time=time(NULL);
	
	/*cerr<<"Printing marginals\n";
	for(int i=0;i<gm.variables.size();i++)
	{
		vector<Variable*> curr_variable;
		curr_variable.push_back(gm.variables[i]);
		Function marginal;
		for(int j=0;j<jg.nodes.size();j++)
		{
			if(do_set_inclusion(curr_variable,jg.nodes[j]->variables(),less_than_comparator_variable))
			{
				jg.nodes[j]->getMarginal(curr_variable,marginal);
				break;
			}
		}
		vector<Double>& marg_table=marginal.table();
		cout<<"Marginal of "<<i<<" is ";
		for(int i=0;i<marg_table.size();i++)
		{
			cout<<marg_table[i]<<" ";
		}
		cout<<endl;
	}
	cerr<<"Time required = "<<(end_time-start_time)<<endl;*/
	//IJGP_Sampler sampler;
	//Double pe//=sampler.computePE(gm,jg,0,order,5,-1);
	//;
	

	SampleSearch ss(gm,jg,0,order);
	Double pe=ss.computePE(time_limit,num_samples);
	pe*=gm.mult_factor;
	cout<<"---------------------------RESULT------------------------------\n";
	cout<<"\t Log probability of evidence = "<<-log(pe.value())<<endl;
	cout<<"\t Probability of evidence = "<<Double((double)1.0/pe.value())<<endl;
	cout<<"---------------------------RESULT------------------------------\n";
}