#include <iostream>
#include <fstream>
#include "BE.h"
#include "GM.h"
#include "Graph.h"
#include "RB_IJGP_Sampler.h"
#include "SampleSearch.h"
#include "JT.h"

using namespace std;
#define IBOUNDLIMIT 32768

void computeIBound(GM& gm, int& i_bound)
{
	int max_domain_size=0;
	for(int i=0;i<gm.variables.size();i++)
	{
		if(max_domain_size < (int) gm.variables[i]->domain_size())
			max_domain_size=gm.variables[i]->domain_size();
	}
	cerr<<"Max dom size = "<<max_domain_size<<endl;
	double m=log((double)IBOUNDLIMIT)/log((double)max_domain_size);
	i_bound=(int)m;
}

int main(int argc, char* argv[])
{
	int i_bound=6,p_bound=0,w_bound=0,num_samples=1000000,time_limit=10,num_iterations=10;
	GM gm;
	gm.readUAI08(argv[1]);
	
	// Random Generation: Start
	myRandom::seed=static_cast<unsigned int> (atoi(argv[3]));
	//myRandom::seed=atoi(argv[3]);
	myRandom::base.seed((int)myRandom::seed);
	myRandom::genDoubleVal= RealGenerator(myRandom::base,myRandom::real_dist);
	myRandom::genIntVal=IntGenerator(myRandom::base,myRandom::int_dist);
	
	vector<int> order;
	//Read evidence
	ifstream in(argv[2]);
	int num_evidence=0;
	vector<int> evidence;
	if(in.good())
	{
		in>>num_evidence;
		for(int i=0;i<num_evidence;i++)
		{
			int var,val;
			in>>var>>val;
			gm.variables[var]->value()=val;
			evidence.push_back(var);
		}
	}
	vector<set<int> > clusters;
	double estimate;
	double limit=(double)10000;
	//sscanf(argv[4],"%lf",&limit);
	//sscanf(argv[5],"%d",&i_bound);
	bool run_samplesearch=false;
	//gm.removeIrrelevantNetwork(evidence);
	gm.setEvidenceBeliefsUAI08(evidence);
	if(gm.mode == DET)
	{
		cerr<<"Reducing domains\n";
		gm.reduceDomains();
	}
	computeIBound(gm,i_bound);
	if(gm.mode == DET)
	{
		GM csp;
		gm.convertToCN(csp);
		Graph g;
		g.makeGraph(&csp);
		int treewidth = g.getMinFillOrdering(order);
		cerr<<"I-bound = "<<i_bound<<" Treewidth = "<<treewidth<<endl;
		//if(treewidth >= i_bound)
			run_samplesearch=true;
			//else
			//gm.getClusters(order,clusters);
	}
	if(run_samplesearch || gm.mode == POSITIVE)
	{
		order=vector<int>();
		clusters=vector<set<int> > ();
		estimate=0.0;
		gm.getMinFillOrdering(order,clusters,estimate);
	}
	
		
	JT jt(gm.variables,gm.functions,clusters,order);
	jt.propagate();
	gm.printMarginalsUAI08(jt.marginals);
	
}
