#include "GM.h"
#include "JG.h"
#include "IJGP_Sampler.h"
#include "SampleSearch.h"
#include "Graph.h"
#include <cmath>
#include <cstdlib>
#include <getopt.h>

using namespace std;
/*************************************************************************************/
#ifdef _MSC_VER
#include <ctime>

static inline double cpuTime(void) {
    return (double)clock() / CLOCKS_PER_SEC; }
#else

#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

static inline double cpuTime(void) {
    struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return (double)ru.ru_utime.tv_sec + (double)ru.ru_utime.tv_usec / 1000000; }
#endif

typedef enum
{
	MINFILL,TOPOLOGICAL,MINDEGREE
}ORDERING;



static struct option long_options[]=
{
	{"ordering",required_argument,NULL,'o'},
	{"mode",required_argument,NULL,'o'},
	{"i-bound",required_argument,NULL,'i'},
	{"num-iterations",required_argument,NULL,'j'},
	{"input",required_argument,NULL,'f'},
	{"outfile",required_argument,NULL,'g'},
	{"help", no_argument,NULL,'h'},
	{"evidencefile", no_argument,NULL,'e'},
};

static const char OPTION_STRING[]="o:m:i:j:f:hg:e:";

void print_help()
{
	cout<<"ijgp [options] -f inputfile -e evidencefile \n";
	cout<<" --mode <LS,SS,LSS>: which propagation mode to use\n";
	cout<<" --ordering <minfill,topological,mindegree> : What ordering to use\n";
	cout<<" --i-bound <int> : An integer value for the i-bound\n";
	cout<<" --num-iterations<int> : The maximum number of iterations for which IJGP is run\n";
	cout<<" --outfile <string> : The name of the output file (default is out)\n";
	cout<<" --help : Print this help\n";



}
//unsigned int myRandom::seed=100L;
int main(int argc, char *argv[])
{ 
	int i_bound=3,p_bound=0,w_bound=0,num_samples=-1,time_limit=10,num_iterations=10;
	char mode[100]="SS";
	GM gm;
	char outfilename[100]="out";
	char ordering[100];
	bool writegraph=false;
	if(argc==1)
	{
		print_help();
		return 1;
	}
	while(1)
	{

		int option_index =0;
		int opt = getopt_long(argc,argv,OPTION_STRING,long_options,&option_index);
		// Detect end of options
		if (opt == -1)
			break;
		switch(opt)
		{
		case 'o':
			sscanf(optarg,"%s",&ordering);
			break;
		case 'i':
			sscanf(optarg,"%ld",&i_bound);
			break;
		case 'm':
			strcpy(mode,optarg);
			break;
		case 'j':
			sscanf(optarg,"%ld",&num_iterations);
			break;
		case 'f':
			//if(!gm.readCNF(optarg))
			//{
				gm=GM();
				//gm.readVIB(optarg);
				gm.readErgo(optarg);
				//char evidfile[100];
				//sprintf(evidfile,"%s.evid",optarg);
				//gm.readEvidence(evidfile);
				//gm.readEvidenceErgo(evidfile);
			//}
				break;
		case 'e':
			
				gm.readEvidenceErgo(optarg);
			
			break;
		case 'g':
			strcpy(outfilename,optarg);

			break;
		case 'h':
			print_help();
			return 1;
		default:
			print_help();
			return 1;
		}
	}
	JG_TYPE prop_type=SS;
	cout<<"Mode is "<<mode<<endl;
	if(strcmp(mode,"SS")==0)
		prop_type=SS;
	else if(strcmp(mode,"LSS")==0)
		prop_type=LSS;
	else if (strcmp(mode,"LS")==0)
		prop_type=LS;
	cout<<"Prop type is "<<prop_type<<endl;

	vector<int> order(gm.variables.size());
	if(strcmp(ordering,"minfill")==0)
	{
		Graph g;
		g.makeGraph(&gm);
		g.getMinFillOrdering(order);
	}
	else if(strcmp(ordering,"topological")==0)
	{
		vector<int> sampling_order (order.size());
		//Graph g;
		//g.makeDirectedGraph(&gm);
		//g.getTopologicalOrdering(sampling_order);
		for(int i=0;i<sampling_order.size();i++)
		{
			sampling_order[i]=i;
		}
		for(int i=0;i<sampling_order.size();i++)
		{
			order[i]=sampling_order[order.size()-i-1];
		}
	}
	else //if (strcmp(ordering,"rcordering")==0)
	{
		GM csp;
		gm.convertToCN(csp);
		Graph g;
		g.makeGraph(&csp);
		g.getMinFillOrdering(order);
	}
	cerr<<"i-bound= "<<i_bound<<endl;
	JG jg(gm,i_bound,num_iterations,order,prop_type);
	double time=cpuTime();
	jg.propagate();
	time-=cpuTime();
	vector<vector<Double> > beliefs(gm.variables.size());
	//Find the node in which the variable occurs
	for(int i=0;i<gm.variables.size();i++)
	{
		int curr_node=-1;
		for(int j=0;j<jg.nodes.size();j++)
		{
			for(int k=0;k<jg.nodes[j]->variables().size();k++)
			{
				if(jg.nodes[j]->variables()[k]->id()==i)
				{

					curr_node=j;
					break;
				}
			}
			if(curr_node > -1)
				break;
		}
		if(curr_node > -1)
		{
			Function function;
			vector<Variable*> marg_variables;
			marg_variables.push_back(gm.variables[i]);
			jg.nodes[curr_node]->getMarginal(marg_variables, function);
			beliefs[i]=function.table();
		}
		else
		{
			beliefs[i]=vector<Double> (gm.variables[i]->domain_size());
			for(int j=0;j<beliefs[i].size();j++)
				beliefs[i][j]=Double((double)1.0/(double) gm.variables[i]->domain_size());
		}
	}
	ofstream outfile(outfilename);
	outfile<<beliefs.size()<<endl;
	for(int i=0;i<beliefs.size();i++)
	{
		outfile<<gm.var_to_ids[i]<<" "<<beliefs[i].size()<<" ";
		for(int j=0;j<beliefs[i].size();j++)
		{
			outfile<<beliefs[i][j]<<" ";
		}
		outfile<<endl;
	}
	outfile<<time<<endl;
	cerr<<"Time  = "<<time<<endl;
	outfile.close();
}