#include <iostream>
#include <fstream>
#include "BE.h"
#include "GM.h"
#include "Graph.h"
#include "RB_IJGP_Sampler.h"
#include "SampleSearch.h"
#include "JT.h"

using namespace std;
#define IBOUNDLIMIT 32768
#include <sys/time.h>
#include <sys/resource.h>
static inline double get_cpu_time(void) 
{
  struct rusage ru;
  getrusage(RUSAGE_SELF, &ru);
  return (double)ru.ru_utime.tv_sec + (double)ru.ru_utime.tv_usec / 1000000; 
}
void computeIBound(GM& gm, int& i_bound)
{
	int max_domain_size=0;
	for(int i=0;i<gm.variables.size();i++)
	{
		if(max_domain_size < (int) gm.variables[i]->domain_size())
			max_domain_size=gm.variables[i]->domain_size();
	}
	cerr<<"Max dom size = "<<max_domain_size<<endl;
	double m=log((double)IBOUNDLIMIT)/log((double)max_domain_size);
	i_bound=(int)m;
}

int main(int argc, char* argv[])
{
	int i_bound=6,p_bound=0,w_bound=0,num_samples=1000000,time_limit=10,num_iterations=10;
	GM gm;
	gm.readUAI08(argv[1]);
	Graph temp_graph;
	vector<int> topo_order;
	if (gm.type == BAYES){
	  temp_graph.makeDirectedGraph(&gm);
	  temp_graph.getTopologicalOrdering(topo_order);
	}
	// Random Generation: Start
	myRandom::seed=static_cast<unsigned int> (atoi(argv[3]));
	//myRandom::seed=atoi(argv[3]);
	myRandom::base.seed((int)myRandom::seed);
	myRandom::genDoubleVal= RealGenerator(myRandom::base,myRandom::real_dist);
	myRandom::genIntVal=IntGenerator(myRandom::base,myRandom::int_dist);
	
	vector<int> order;
	//Read evidence
	ifstream in(argv[2]);
	int num_evidence=0;
	vector<int> evidence;
	if(in.good())
	{
		in>>num_evidence;
		for(int i=0;i<num_evidence;i++)
		{
			int var,val;
			in>>var>>val;
			gm.variables[var]->value()=val;
			evidence.push_back(var);
		}
	}
	vector<set<int> > clusters;
	double estimate;
	double limit=(double)1000;
	//sscanf(argv[4],"%lf",&limit);
	//sscanf(argv[5],"%d",&i_bound);
	bool run_samplesearch=false;
	//gm.removeIrrelevantNetwork(evidence);
	gm.setEvidenceBeliefsUAI08(evidence);
	if(gm.mode == DET)
	{
		cerr<<"Reducing domains\n";
		gm.reduceDomains();
	}
	computeIBound(gm,i_bound);
	if(gm.mode == DET)
	{
		GM csp;
		gm.convertToCN(csp);
		Graph g;
		g.makeGraph(&csp);
		int treewidth = g.getMinFillOrdering(order);
		cerr<<"I-bound = "<<i_bound<<" Treewidth = "<<treewidth<<endl;
		//if(treewidth >= i_bound)
			run_samplesearch=true;
			//else
			//gm.getClusters(order,clusters);
	}
	if(run_samplesearch || gm.mode == POSITIVE)
	{
		order=vector<int>();
		clusters=vector<set<int> > ();
		estimate=0.0;
		
		if (gm.type == BAYES){
		  if (topo_order.empty()){
		    gm.getMinFillOrdering(order,clusters,estimate);
		  }
		  else{
		    cerr<<"Topo order\n";
		    order=vector<int>();
		    map<int,int> old2new;
		    for(int i=0;i<gm.variables.size();i++){
		      old2new[gm.variables[i]->orig_id]=gm.variables[i]->id();
		    }
		    for(int i=topo_order.size()-1;i>-1;i--){
		      if(old2new.find(topo_order[i])!=old2new.end()){
			order.push_back(old2new[topo_order[i]]);
		      }
		    }
		    assert(order.size()==gm.variables.size());
		    gm.getClusters(order,clusters);
		  }
		}
		else{
		  gm.getMinFillOrdering(order,clusters,estimate);
		}
	}
	
	cerr<<"num evidence = "<<num_evidence<<endl;

	
	//JT jt(gm.variables,gm.functions,clusters,order);
	//jt.propagate();
	//gm.printMarginalsUAI08(jt.marginals);
	//Graph graph;
	//graph.makeGraph(&gm);
	//graph.getMinFillOrdering(order);
	
	//
	//BE be(gm.variables,gm.functions,order);

	
	//for(int i=0;i<order.size();i++)
	//	cerr<<order[i]<<" ";
	//cerr<<endl;
	//
	//cerr<<"Limit = "<<limit<<endl;
	//vector<int> new_order;
	//gm.rearrangeOrdering(order,clusters,new_order,limit);
	//cerr<<"#sampled variables = "<<new_order.size()<<" out of "<<gm.variables.size()<<endl;
	////BE be(gm.variables,gm.functions,order);

	////be.log_pe+=LogDouble(gm.mult_factor);
	////cerr<<"Estimate = "<<estimate<<endl;
	////cout<<"LogPe = "<<be.log_pe.val<<endl;


	//cout<<order.size()<<endl;
	//JG jg(gm,i_bound,num_iterations,order,LSS);

	//cerr<<"Join graph constructed\n";
	//
	//
	//jg.propagate();
	//

	//if(!run_samplesearch)
	//{
	//	RB_IJGP_Sampler sampler;
	//	Double pe=sampler.computePE(gm,jg,p_bound,order,new_order,time_limit,num_samples);
	//	pe*=gm.mult_factor;

	//	cerr<<"---------------------------RESULT------------------------------\n";
	//	cerr<<"\t Log Base e probability of evidence = "<<log(pe.value())<<endl;
	//	cerr<<"\t Log Base 10 probability of evidence = "<<log(pe.value())/log((double)10.0)<<endl;
	//	cerr<<"\t Probability of evidence = "<<pe<<endl;
	//	cerr<<"---------------------------RESULT------------------------------\n";
	//}
	//else
	//{
	//	cerr<<"Running SampleSarch\n";
	//	RBSampleSearch ss;
	//	Double pe1=ss.computePE(gm,jg,p_bound,order,new_order,num_samples);
	//	pe1*=gm.mult_factor;
	//	cerr<<"---------------------------RESULT------------------------------\n";
	//	cerr<<"\t Log Base e probability of evidence = "<<log(pe1.value())<<endl;
	//	cerr<<"\t Log Base 10 probability of evidence = "<<log(pe1.value())/log((double)10.0)<<endl;
	//	cerr<<"\t Probability of evidence = "<<pe1<<endl;
	//	cerr<<"---------------------------RESULT------------------------------\n";
	//}
	cout <<"t "<<get_cpu_time()<<endl;
	JG jg(gm,i_bound,num_iterations,order,LSS);
	
	jg.propagate();
	
	vector<int> new_order;
	//for(int i=0;i<order.size();i++)
	//	new_order.push_back(order[order.size()-i-1]);
	gm.rearrangeOrdering(order,clusters,new_order,limit);
	cerr<<"#Sampled variables = "<<new_order.size()<<"  out of "<<order.size()<<endl;
	run_samplesearch=false;
	if(!run_samplesearch)
	{
		RB_IJGP_Sampler sampler;
		sampler.computeBeliefs(gm,jg,p_bound,clusters,order,new_order,time_limit,num_samples);
	}
	else
	{
		RBSampleSearch ss;
		ss.computeBeliefs(gm,jg,p_bound,clusters,order,new_order,num_samples);
	}
}
